#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import *
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "5.2"

# this is the default in gen/plug/_pluginreg.py: plg.require_active = True

# ------------------------------------------------------------------------
#
# Ancestor Tree
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "ancestor_chart,BKI"
plg.name = _("Ancestor Chart")
plg.description = _("Produces a graphical ancestral chart")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "ancestortree.py"
plg.ptype = REPORT
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.category = CATEGORY_DRAW
plg.reportclass = "AncestorTree"
plg.optionclass = "AncestorTreeOptions"
plg.report_modes = [REPORT_MODE_BKI]

plg = newplugin()
plg.id = "ancestor_chart"
plg.name = _("Ancestor Tree")
plg.description = _("Produces a graphical ancestral tree")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "ancestortree.py"
plg.ptype = REPORT
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.category = CATEGORY_DRAW
plg.reportclass = "AncestorTree"
plg.optionclass = "AncestorTreeOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]

# ------------------------------------------------------------------------
#
# Calendar
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "calendar"
plg.name = _("Calendar")
plg.description = _("Produces a graphical calendar")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "calendarreport.py"
plg.ptype = REPORT
plg.authors = ["Douglas S. Blank"]
plg.authors_email = ["doug.blank@gmail.com"]
plg.category = CATEGORY_DRAW
plg.reportclass = "Calendar"
plg.optionclass = "CalendarOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]

# ------------------------------------------------------------------------
#
# Descendant Tree
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "descend_chart,BKI"
plg.name = _("Descendant Chart")
plg.description = _("Produces a graphical descendant chart")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "descendtree.py"
plg.ptype = REPORT
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.category = CATEGORY_DRAW
plg.reportclass = "DescendTree"
plg.optionclass = "DescendTreeOptions"
plg.report_modes = [REPORT_MODE_BKI]

plg = newplugin()
plg.id = "descend_chart"
plg.name = _("Descendant Tree")
plg.description = _("Produces a graphical descendant tree")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "descendtree.py"
plg.ptype = REPORT
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.category = CATEGORY_DRAW
plg.reportclass = "DescendTree"
plg.optionclass = "DescendTreeOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]

# ------------------------------------------------------------------------
#
# Family Descendant Tree
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "family_descend_chart,BKI"
plg.name = _("Family Descendant Chart")
plg.description = _("Produces a graphical descendant chart around a family")
plg.version = "1.0"
plg.status = STABLE
plg.fname = "descendtree.py"
plg.ptype = REPORT
plg.category = CATEGORY_DRAW
plg.gramps_target_version = MODULE_VERSION
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.require_active = True
plg.reportclass = "DescendTree"
plg.optionclass = "DescendTreeOptions"
plg.report_modes = [REPORT_MODE_BKI]

plg = newplugin()
plg.id = "family_descend_chart"
plg.name = _("Family Descendant Tree")
plg.description = _("Produces a graphical descendant tree around a family")
plg.version = "1.0"
plg.status = STABLE
plg.fname = "descendtree.py"
plg.ptype = REPORT
plg.category = CATEGORY_DRAW
plg.gramps_target_version = MODULE_VERSION
plg.authors = ["Craig J. Anderson"]
plg.authors_email = ["ander882@hotmail.com"]
plg.require_active = True
plg.reportclass = "DescendTree"
plg.optionclass = "DescendTreeOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]

# ------------------------------------------------------------------------
#
# Fan Chart
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "fan_chart"
plg.name = _("Fan Chart")
plg.description = _("Produces fan charts")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "fanchart.py"
plg.ptype = REPORT
plg.authors = ["Donald N. Allingham"]
plg.authors_email = ["don@gramps-project.org"]
plg.category = CATEGORY_DRAW
plg.reportclass = "FanChart"
plg.optionclass = "FanChartOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]

# ------------------------------------------------------------------------
#
# Statistics Charts
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "statistics_chart"
plg.name = _("Statistics Charts")
plg.description = _(
    "Produces statistical bar and pie charts of the people " "in the database"
)
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "statisticschart.py"
plg.ptype = REPORT
plg.authors = ["Eero Tamminen"]
plg.authors_email = [""]
plg.category = CATEGORY_DRAW
plg.reportclass = "StatisticsChart"
plg.optionclass = "StatisticsChartOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]
plg.require_active = False

# ------------------------------------------------------------------------
#
# Timeline Chart
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "timeline"
plg.name = _("Timeline Chart")
plg.description = _("Produces a timeline chart.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "timeline.py"
plg.ptype = REPORT
plg.authors = ["Donald N. Allingham"]
plg.authors_email = ["don@gramps-project.org"]
plg.category = CATEGORY_DRAW
plg.reportclass = "TimeLine"
plg.optionclass = "TimeLineOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._hassourceofbase import HasSourceOfBase


# -------------------------------------------------------------------------
#
# HasSourceOf
#
# -------------------------------------------------------------------------
class HasSourceOf(HasSourceOfBase):
    """Rule that checks media that have a particular source."""

    labels = [_("Source ID:")]
    name = _("Media with the <source>")
    category = _("Citation/source filters")
    description = _("Matches media who have a particular source")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
#

"""
Display filtered data
"""

from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gui.plug.quick import QuickTable
from gramps.gen.utils.file import media_path_full
from gramps.gui.plug.quick import run_quick_report_by_name_direct
from gramps.gen.lib import Person
from gramps.gen.datehandler import get_date

import os
from collections import defaultdict
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored

fname_map = {
    "all": _("all", "Filtering_on"),
    "Inverse Person": _("Inverse Person", "Filtering_on"),
    "Inverse Family": _("Inverse Family", "Filtering_on"),
    "Inverse Event": _("Inverse Event", "Filtering_on"),
    "Inverse Place": _("Inverse Place", "Filtering_on"),
    "Inverse Source": _("Inverse Source", "Filtering_on"),
    "Inverse Repository": _("Inverse Repository", "Filtering_on"),
    "Inverse Media": _("Inverse Media", "Filtering_on"),
    "Inverse Note": _("Inverse Note", "Filtering_on"),
    "all people": _("all people", "Filtering_on"),
    "all families": _("all families", "Filtering_on"),
    "all events": _("all events", "Filtering_on"),
    "all places": _("all places", "Filtering_on"),
    "all sources": _("all sources", "Filtering_on"),
    "all repositories": _("all repositories", "Filtering_on"),
    "all media": _("all media", "Filtering_on"),
    "all notes": _("all notes", "Filtering_on"),
    "males": _("males", "Filtering_on"),
    "females": _("females", "Filtering_on"),
    "people with other gender": _("people with other gender", "Filtering_on"),
    "people with unknown gender": _("people with unknown gender", "Filtering_on"),
    "incomplete names": _("incomplete names", "Filtering_on"),
    "people with missing birth dates": _(
        "people with missing birth dates", "Filtering_on"
    ),
    "disconnected people": _("disconnected people", "Filtering_on"),
    "unique surnames": _("unique surnames", "Filtering_on"),
    "people with media": _("people with media", "Filtering_on"),
    "media references": _("media references", "Filtering_on"),
    "unique media": _("unique media", "Filtering_on"),
    "missing media": _("missing media", "Filtering_on"),
    "media by size": _("media by size", "Filtering_on"),
    "list of people": _("list of people", "Filtering_on"),
}


def run(database, document, filter_name, *args, **kwargs):
    """
    Loops through the families that the person is a child in, and display
    the information about the other children.
    """
    # setup the simple access functions
    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)
    if filter_name == "all":
        sdoc.title(_("Summary counts of current selection"))
        sdoc.paragraph("")
        sdoc.paragraph(_("Right-click row (or press ENTER) to see selected items."))
        sdoc.paragraph("")
        stab.columns(_("Object"), _("Count/Total"))
        if hasattr(database, "db"):
            stab.row(
                [_("People"), "Filter", "Person"],
                "%d/%d"
                % (
                    len(database.get_person_handles()),
                    len(database.db.get_person_handles()),
                ),
            )
            stab.row(
                [_("Families"), "Filter", "Family"],
                "%d/%d"
                % (
                    len(database.get_family_handles()),
                    len(database.db.get_family_handles()),
                ),
            )
            stab.row(
                [_("Events"), "Filter", "Event"],
                "%d/%d"
                % (
                    len(database.get_event_handles()),
                    len(database.db.get_event_handles()),
                ),
            )
            stab.row(
                [_("Places"), "Filter", "Place"],
                "%d/%d"
                % (
                    len(database.get_place_handles()),
                    len(database.db.get_place_handles()),
                ),
            )
            stab.row(
                [_("Sources"), "Filter", "Source"],
                "%d/%d"
                % (
                    len(database.get_source_handles()),
                    len(database.db.get_source_handles()),
                ),
            )
            stab.row(
                [_("Repositories"), "Filter", "Repository"],
                "%d/%d"
                % (
                    len(database.get_repository_handles()),
                    len(database.db.get_repository_handles()),
                ),
            )
            stab.row(
                [_("Media"), "Filter", "Media"],
                "%d/%d"
                % (
                    len(database.get_media_handles()),
                    len(database.db.get_media_handles()),
                ),
            )
            stab.row(
                [_("Notes"), "Filter", "Note"],
                "%d/%d"
                % (
                    len(database.get_note_handles()),
                    len(database.db.get_note_handles()),
                ),
            )
        else:
            stab.row(
                [_("People"), "Filter", "Person"],
                "%d/%d"
                % (
                    len(database.get_person_handles()),
                    len(database.basedb.get_person_handles()),
                ),
            )
            stab.row(
                [_("Families"), "Filter", "Family"],
                "%d/%d"
                % (
                    len(database.get_family_handles()),
                    len(database.basedb.get_family_handles()),
                ),
            )
            stab.row(
                [_("Events"), "Filter", "Event"],
                "%d/%d"
                % (
                    len(database.get_event_handles()),
                    len(database.basedb.get_event_handles()),
                ),
            )
            stab.row(
                [_("Places"), "Filter", "Place"],
                "%d/%d"
                % (
                    len(database.get_place_handles()),
                    len(database.basedb.get_place_handles()),
                ),
            )
            stab.row(
                [_("Sources"), "Filter", "Source"],
                "%d/%d"
                % (
                    len(database.get_source_handles()),
                    len(database.basedb.get_source_handles()),
                ),
            )
            stab.row(
                [_("Repositories"), "Filter", "Repository"],
                "%d/%d"
                % (
                    len(database.get_repository_handles()),
                    len(database.basedb.get_repository_handles()),
                ),
            )
            stab.row(
                [_("Media"), "Filter", "Media"],
                "%d/%d"
                % (
                    len(database.get_media_handles()),
                    len(database.basedb.get_media_handles()),
                ),
            )
            stab.row(
                [_("Notes"), "Filter", "Note"],
                "%d/%d"
                % (
                    len(database.get_note_handles()),
                    len(database.basedb.get_note_handles()),
                ),
            )
        sdoc.paragraph("")
        stab.write(sdoc)
        return

    # display the title
    if filter_name in fname_map:
        sdoc.title(_("Filtering on %s") % fname_map[filter_name])  # listed above
    else:
        sdoc.title(_("Filtering on %s") % _(filter_name))
    sdoc.paragraph("")
    matches = 0

    if filter_name == "Inverse Person":
        sdb.dbase = database.db
        stab.columns(_("Person"), _("Gramps ID"), _("Birth Date"))
        proxy_handles = set(database.iter_person_handles())

        for person in database.db.iter_people():
            if person.handle not in proxy_handles:
                stab.row(person, person.gramps_id, sdb.birth_or_fallback(person))
                matches += 1

    elif filter_name == "Inverse Family":
        sdb.dbase = database.db
        stab.columns(_("Family"), _("Gramps ID"))
        proxy_handles = set(database.iter_family_handles())

        for family in database.db.iter_families():
            if family.handle not in proxy_handles:
                stab.row(family, family.gramps_id)
                matches += 1

    elif filter_name == "Inverse Event":
        sdb.dbase = database.db
        stab.columns(_("Event"), _("Gramps ID"))
        proxy_handles = set(database.iter_event_handles())

        for event in database.db.iter_events():
            if event.handle not in proxy_handles:
                stab.row(event, event.gramps_id)
                matches += 1

    elif filter_name == "Inverse Place":
        sdb.dbase = database.db
        stab.columns(_("Place"), _("Gramps ID"))
        proxy_handles = set(database.iter_place_handles())

        for place in database.db.iter_places():
            if place.handle not in proxy_handles:
                stab.row(place, place.gramps_id)
                matches += 1

    elif filter_name == "Inverse Source":
        sdb.dbase = database.db
        stab.columns(_("Source"), _("Gramps ID"))
        proxy_handles = set(database.iter_source_handles())

        for source in database.db.iter_sources():
            if source.handle not in proxy_handles:
                stab.row(source, source.gramps_id)
                matches += 1

    elif filter_name == "Inverse Repository":
        sdb.dbase = database.db
        stab.columns(_("Repository"), _("Gramps ID"))
        proxy_handles = set(database.iter_repository_handles())

        for repository in database.db.iter_repositories():
            if repository.handle not in proxy_handles:
                stab.row(repository, repository.gramps_id)
                matches += 1

    elif filter_name == "Inverse Media":
        sdb.dbase = database.db
        stab.columns(_("Media"), _("Gramps ID"))
        proxy_handles = set(database.iter_media_handles())

        for media in database.db.iter_media():
            if media.handle not in proxy_handles:
                stab.row(media, media.gramps_id)
                matches += 1

    elif filter_name == "Inverse Note":
        sdb.dbase = database.db
        stab.columns(_("Note"), _("Gramps ID"))
        proxy_handles = set(database.iter_note_handles())

        for note in database.db.iter_notes():
            if note.handle not in proxy_handles:
                stab.row(note, note.gramps_id)
                matches += 1

    elif filter_name in ["all people", "Person"]:
        stab.columns(_("Person"), _("Gramps ID"), _("Birth Date"))
        for person in database.iter_people():
            stab.row(person, person.gramps_id, sdb.birth_or_fallback(person))
            matches += 1

    elif filter_name in ["all families", "Family"]:
        stab.columns(_("Family"), _("Gramps ID"))
        for family in database.iter_families():
            stab.row(family, family.gramps_id)
            matches += 1

    elif filter_name in ["all events", "Event"]:
        stab.columns(_("Event"), _("Gramps ID"))
        for obj in database.iter_events():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name in ["all places", "Place"]:
        stab.columns(_("Place"), _("Gramps ID"))
        for obj in database.iter_places():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name in ["all sources", "Source"]:
        stab.columns(_("Source"), _("Gramps ID"))
        for obj in database.iter_sources():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name in ["all repositories", "Repository"]:
        stab.columns(_("Repository"), _("Gramps ID"))
        for obj in database.iter_repositories():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name in ["all media", "Media"]:
        stab.columns(_("Media"), _("Gramps ID"))
        for obj in database.iter_media():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name in ["all notes", "Note"]:
        stab.columns(_("Note"), _("Gramps ID"))
        for obj in database.iter_notes():
            stab.row(obj, obj.gramps_id)
            matches += 1

    elif filter_name == "males":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            if person.gender == Person.MALE:
                stab.row(
                    person,
                    sdb.birth_or_fallback(person),
                    str(person.get_primary_name().get_type()),
                )
                matches += 1

    elif filter_name == "females":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            if person.gender == Person.FEMALE:
                stab.row(
                    person,
                    sdb.birth_or_fallback(person),
                    str(person.get_primary_name().get_type()),
                )
                matches += 1

    elif filter_name == "people with other gender":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            if person.gender == Person.OTHER:
                stab.row(
                    person,
                    sdb.birth_or_fallback(person),
                    str(person.get_primary_name().get_type()),
                )
                matches += 1

    elif filter_name == "people with unknown gender":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            if person.gender == Person.UNKNOWN:
                stab.row(
                    person,
                    sdb.birth_or_fallback(person),
                    str(person.get_primary_name().get_type()),
                )
                matches += 1

    elif filter_name == "incomplete names":
        stab.columns(_("Name"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            for name in [person.get_primary_name()] + person.get_alternate_names():
                if name.get_first_name().strip() == "":
                    stab.row(
                        [name.get_name(), "Person", person.handle],
                        sdb.birth_or_fallback(person),
                        str(name.get_type()),
                    )
                    matches += 1
                else:
                    if name.get_surname_list():
                        for surname in name.get_surname_list():
                            if surname.get_surname().strip() == "":
                                stab.row(
                                    [name.get_first_name(), "Person", person.handle],
                                    sdb.birth_or_fallback(person),
                                    str(name.get_type()),
                                )
                                matches += 1
                    else:
                        stab.row(
                            [name.get_first_name(), "Person", person.handle],
                            sdb.birth_or_fallback(person),
                            str(name.get_type()),
                        )
                        matches += 1

    elif filter_name == "people with missing birth dates":
        stab.columns(_("Person"), _("Type"))
        for person in database.iter_people():
            birth_ref = person.get_birth_ref()
            if birth_ref:
                birth = database.get_event_from_handle(birth_ref.ref)
                if not get_date(birth):
                    stab.row(person, _("birth event but no date"))
                    matches += 1
            else:
                stab.row(person, _("missing birth event"))
                matches += 1

    elif filter_name == "disconnected people":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        for person in database.iter_people():
            if (not person.get_main_parents_family_handle()) and (
                not len(person.get_family_handle_list())
            ):
                stab.row(
                    person,
                    sdb.birth_or_fallback(person),
                    str(person.get_primary_name().get_type()),
                )
                matches += 1

    elif filter_name == "unique surnames":
        namelist = defaultdict(int)
        for person in database.iter_people():
            names = [person.get_primary_name()] + person.get_alternate_names()
            surnames = list(set([name.get_group_name() for name in names]))
            for surname in surnames:
                namelist[surname] += 1
        stab.columns(_("Surname"), _("Count"))
        for name in sorted(namelist):
            stab.row(name, namelist[name])
            matches += 1
        stab.set_callback(
            "leftdouble",
            lambda name: run_quick_report_by_name_direct(
                "samesurnames", database, document, name
            ),
        )

    elif filter_name == "people with media":
        stab.columns(_("Person"), _("Media count"))
        for person in database.iter_people():
            length = len(person.get_media_list())
            if length > 0:
                stab.row(person, str(length))
                matches += 1

    elif filter_name == "media references":
        stab.columns(_("Person"), _("Reference"))
        for person in database.iter_people():
            medialist = person.get_media_list()
            for item in medialist:
                stab.row(person, _("media"))
                matches += 1

    elif filter_name == "unique media":
        stab.columns(_("Unique Media"))
        for photo in database.iter_media():
            fullname = media_path_full(database, photo.get_path())
            stab.row(fullname)
            matches += 1

    elif filter_name == "missing media":
        stab.columns(_("Missing Media"))
        for photo in database.iter_media():
            fullname = media_path_full(database, photo.get_path())
            try:
                os.path.getsize(fullname)
            except:
                stab.row(fullname)
                matches += 1

    elif filter_name == "media by size":
        stab.columns(_("Media"), _("Size in bytes"))
        for photo in database.iter_media():
            fullname = media_path_full(database, photo.get_path())
            try:
                bytes = os.path.getsize(fullname)
                stab.row(fullname, str(bytes))
                matches += 1
            except:
                pass

    elif filter_name == "list of people":
        stab.columns(_("Person"), _("Birth Date"), _("Name type"))
        handles = kwargs["handles"]
        for person_handle in handles:
            person = database.get_person_from_handle(person_handle)
            stab.row(
                person,
                sdb.birth_or_fallback(person),
                str(person.get_primary_name().get_type()),
            )
            matches += 1

    else:
        raise AttributeError("invalid filter name: '%s'" % filter_name)
    # Translators: leave all/any {...} untranslated
    sdoc.paragraph(
        ngettext(
            "Filter matched {number_of} record.",
            "Filter matched {number_of} records.",
            matches,
        ).format(number_of=matches)
    )
    sdoc.paragraph("")
    document.has_data = matches > 0
    if matches > 0:
        stab.write(sdoc)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006, 2008  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Export to Web Family Tree"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging
from collections import abc

log = logging.getLogger(".WriteFtree")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
# keep the following line even though not obviously used (works on import)
from gramps.gui.plug.export import WriterOptionBox
from gramps.gui.dialog import ErrorDialog
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# writeData
#
# -------------------------------------------------------------------------
def writeData(database, filename, user, option_box=None):
    """function to export Web Family Tree file"""
    writer = FtreeWriter(database, filename, user, option_box)
    return writer.export_data()


# -------------------------------------------------------------------------
#
# FtreeWriter
#
# -------------------------------------------------------------------------
class FtreeWriter:
    """Export a Web Family Tree format file"""

    def __init__(self, database, filename, user, option_box=None):
        self.db = database
        self.filename = filename
        self.user = user
        self.option_box = option_box
        # is callback is really callable?
        if isinstance(self.user.callback, abc.Callable):
            self.update = self.update_real
        else:
            self.update = self.update_empty

        if option_box:
            self.option_box.parse_options()
            self.db = option_box.get_filtered_database(self.db)

        self.plist = self.db.get_person_handles()
        self.plist.sort()
        # the following are used to update the progress meter
        self.total = 2 * len(self.plist)
        self.count = 0
        self.oldval = 0  # we only update when percentage changes

    def update_empty(self):
        """used when no callback is present"""
        pass

    def update_real(self):
        """Progress update"""
        self.count += 1
        newval = int(100 * self.count / self.total)
        if newval != self.oldval:
            self.user.callback(newval)
            self.oldval = newval

    def export_data(self):
        """main export processing"""
        name_map = {}
        id_map = {}
        id_name = {}

        for key in self.plist:
            self.update()
            pnam = self.db.get_person_from_handle(key).get_primary_name()
            snam = pnam.get_surname()
            items = pnam.get_first_name().split()
            nam = ("%s %s" % (items[0], snam)) if items else snam

            count = -1
            if nam in name_map:
                count = 0
                while 1:
                    nam_num = "%s%d" % (nam, count)
                    if nam_num not in name_map:
                        break
                    count += 1
                name_map[nam_num] = key
                id_map[key] = nam_num
            else:
                name_map[nam] = key
                id_map[key] = nam
            id_name[key] = get_name(pnam, snam, count)

        try:
            with open(self.filename, "w", encoding="utf_8") as file:
                return self._export_data(file, id_name, id_map)
        except IOError as msg:
            msg2 = _("Could not create %s") % self.filename
            ErrorDialog(msg2, str(msg), parent=self.option_box.window)
            return False

    def _export_data(self, file, id_name, id_map):
        """file export processing"""
        for key in self.plist:
            self.update()
            pers = self.db.get_person_from_handle(key)
            name = id_name[key]
            father = mother = email = web = ""

            family_handle = pers.get_main_parents_family_handle()
            if family_handle:
                family = self.db.get_family_from_handle(family_handle)
                if family.get_father_handle() and family.get_father_handle() in id_map:
                    father = id_map[family.get_father_handle()]
                if family.get_mother_handle() and family.get_mother_handle() in id_map:
                    mother = id_map[family.get_mother_handle()]

            #
            # Calculate Date
            #
            birth_ref = pers.get_birth_ref()
            death_ref = pers.get_death_ref()
            if birth_ref:
                birth_event = self.db.get_event_from_handle(birth_ref.ref)
                birth = birth_event.get_date_object()
            else:
                birth = None
            if death_ref:
                death_event = self.db.get_event_from_handle(death_ref.ref)
                death = death_event.get_date_object()
            else:
                death = None

            # if self.restrict:
            #    alive = probably_alive(pers, self.db)
            # else:
            #    alive = 0

            if birth:
                if death:
                    dates = "%s-%s" % (fdate(birth), fdate(death))
                else:
                    dates = fdate(birth)
            else:
                if death:
                    dates = fdate(death)
                else:
                    dates = ""

            file.write(
                "%s;%s;%s;%s;%s;%s\n" % (name, father, mother, email, web, dates)
            )

        return True


def fdate(val):
    """return properly formatted date"""
    if val.get_year_valid():
        if val.get_month_valid():
            if val.get_day_valid():
                return "%d/%d/%d" % (val.get_day(), val.get_month(), val.get_year())
            return "%d/%d" % (val.get_month(), val.get_year())
        return "%d" % val.get_year()
    return ""


def get_name(name, surname, count):
    """returns a name string built from the components of the Name
    instance, in the form of Firstname Surname"""

    return (
        name.first_name
        + " "
        + surname
        + (str(count) if count != -1 else "")
        + (", " + name.suffix if name.suffix else "")
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Analysis and Exploration/Compare Individual Events"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import os
from collections import defaultdict

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.filters import GenericFilter, rules
from gramps.gui.filters import build_filter_model
from gramps.gen.sort import Sort
from gramps.gui.utils import ProgressMeter
from gramps.gen.utils.docgen import ODSTab
from gramps.gen.const import CUSTOM_FILTERS, URL_MANUAL_PAGE
from gramps.gen.errors import WindowActiveError
from gramps.gen.datehandler import get_date
from gramps.gui.dialog import WarningDialog
from gramps.gui.plug import tool
from gramps.gen.plug.report import utils
from gramps.gui.display import display_help
from gramps.gui.managedwindow import ManagedWindow
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gui.glade import Glade
from gramps.gui.editors import FilterEditor
from gramps.gen.constfunc import get_curr_dir
from gramps.gen.display.place import displayer as _pd

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Compare_Individual_Events", "manual")


# ------------------------------------------------------------------------
#
# EventCmp
#
# ------------------------------------------------------------------------
class TableReport:
    """
    This class provides an interface for the spreadsheet table
    used to save the data into the file.
    """

    def __init__(self, filename, doc):
        self.filename = filename
        self.doc = doc

    def initialize(self, cols):
        self.doc.open(self.filename)
        self.doc.start_page()

    def finalize(self):
        self.doc.end_page()
        self.doc.close()

    def write_table_data(self, data, skip_columns=[]):
        self.doc.start_row()
        index = -1
        for item in data:
            index += 1
            if index not in skip_columns:
                self.doc.write_cell(item)
        self.doc.end_row()

    def set_row(self, val):
        self.row = val + 2

    def write_table_head(self, data):
        self.doc.start_row()
        list(map(self.doc.write_cell, data))
        self.doc.end_row()


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class EventComparison(tool.Tool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.dbstate = dbstate
        self.uistate = uistate

        tool.Tool.__init__(self, dbstate, options_class, name)
        ManagedWindow.__init__(self, uistate, [], self)
        self.qual = 0

        self.filterDialog = Glade(toplevel="filters", also_load=["liststore1"])
        self.filterDialog.connect_signals(
            {
                "on_apply_clicked": self.on_apply_clicked,
                "on_editor_clicked": self.filter_editor_clicked,
                "on_help_clicked": self.on_help_clicked,
                "destroy_passed_object": self.close,
                "on_write_table": self.__dummy,
            }
        )

        window = self.filterDialog.toplevel
        self.filters = self.filterDialog.get_object("filter_list")
        self.label = _("Event comparison filter selection")
        self.set_window(window, self.filterDialog.get_object("title"), self.label)
        self.setup_configs("interface.eventcomparison", 640, 220)

        self.on_filters_changed("Person")
        uistate.connect("filters-changed", self.on_filters_changed)

        self.show()

    def __dummy(self, obj):
        """dummy callback, needed because widget is in same glade file
        as another widget, so callbacks must be defined to avoid warnings.
        """
        pass

    def on_filters_changed(self, name_space):
        if name_space == "Person":
            all_filter = GenericFilter()
            all_filter.set_name(_("Entire Database"))
            all_filter.add_rule(rules.person.Everyone([]))
            self.filter_model = build_filter_model("Person", [all_filter])
            self.filters.set_model(self.filter_model)
            self.filters.set_active(0)

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def build_menu_names(self, obj):
        return (_("Filter selection"), _("Event Comparison tool"))

    def filter_editor_clicked(self, obj):
        try:
            FilterEditor("Person", CUSTOM_FILTERS, self.dbstate, self.uistate)
        except WindowActiveError:
            pass

    def on_apply_clicked(self, obj):
        cfilter = self.filter_model[self.filters.get_active()][1]

        progress_bar = ProgressMeter(_("Comparing events"), "", parent=self.window)
        progress_bar.set_pass(_("Selecting people"), 1)

        plist = cfilter.apply(self.db, self.db.iter_person_handles())

        progress_bar.step()
        progress_bar.close()
        self.options.handler.options_dict["filter"] = self.filters.get_active()
        # Save options
        self.options.handler.save_options()

        if len(plist) == 0:
            WarningDialog(_("No matches were found"), parent=self.window)
        else:
            EventComparisonResults(self.dbstate, self.uistate, plist, self.track)


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
##def by_value(first,second):
##    return cmp(second[0],first[0])


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def fix(line):
    l = line.strip().replace("&", "&amp;").replace(">", "&gt;")
    return l.replace(l, "<", "&lt;").replace(l, '"', "&quot;")


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class EventComparisonResults(ManagedWindow):
    def __init__(self, dbstate, uistate, people_list, track):
        self.dbstate = dbstate
        self.uistate = uistate

        ManagedWindow.__init__(self, uistate, track, self)

        self.db = dbstate.db
        self.my_list = people_list
        self.row_data = []
        self.save_form = None

        self.topDialog = Glade(toplevel="eventcmp")
        self.topDialog.connect_signals(
            {
                "on_write_table": self.on_write_table,
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_apply_clicked": self.__dummy,
                "on_editor_clicked": self.__dummy,
            }
        )

        window = self.topDialog.toplevel
        self.set_window(
            window, self.topDialog.get_object("title"), _("Event Comparison Results")
        )
        self.setup_configs("interface.eventcomparisonresults", 750, 400)

        self.eventlist = self.topDialog.get_object("treeview")
        self.sort = Sort(self.db)
        self.my_list.sort(key=self.sort.by_last_name_key)

        self.event_titles = self.make_event_titles()

        self.table_titles = [_("Person"), _("ID")]
        for event_name in self.event_titles:
            self.table_titles.append(
                _("%(event_name)s Date") % {"event_name": event_name}
            )
            self.table_titles.append("sort")  # This won't be shown in a tree
            self.table_titles.append(
                _("%(event_name)s Place") % {"event_name": event_name}
            )

        self.build_row_data()
        self.draw_display()
        self.show()

    def __dummy(self, obj):
        """dummy callback, needed because widget is in same glade file
        as another widget, so callbacks must be defined to avoid warnings.
        """
        pass

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def build_menu_names(self, obj):
        return (_("Event Comparison Results"), None)

    def draw_display(self):
        model_index = 0
        tree_index = 0
        mylist = []
        renderer = Gtk.CellRendererText()
        for title in self.table_titles:
            mylist.append(str)
            if title == "sort":
                # This will override the previously defined column
                self.eventlist.get_column(tree_index - 1).set_sort_column_id(
                    model_index
                )
            else:
                column = Gtk.TreeViewColumn(title, renderer, text=model_index)
                column.set_sort_column_id(model_index)
                self.eventlist.append_column(column)
                # This one numbers the tree columns: increment on new column
                tree_index += 1
            # This one numbers the model columns: always increment
            model_index += 1

        model = Gtk.ListStore(*mylist)
        self.eventlist.set_model(model)

        self.progress_bar.set_pass(_("Building display"), len(self.row_data))
        for data in self.row_data:
            model.append(row=list(data))
            self.progress_bar.step()
        self.progress_bar.close()

    def build_row_data(self):
        self.progress_bar = ProgressMeter(
            _("Comparing Events"), "", parent=self.uistate.window
        )
        self.progress_bar.set_pass(_("Building data"), len(self.my_list))
        for individual_id in self.my_list:
            individual = self.db.get_person_from_handle(individual_id)
            name = individual.get_primary_name().get_name()
            gid = individual.get_gramps_id()

            the_map = defaultdict(list)
            for ievent_ref in individual.get_event_ref_list():
                ievent = self.db.get_event_from_handle(ievent_ref.ref)
                event_name = str(ievent.get_type())
                the_map[event_name].append(ievent_ref.ref)

            first = True
            done = False
            while not done:
                added = False
                tlist = [name, gid] if first else ["", ""]

                for ename in self.event_titles:
                    if ename in the_map and len(the_map[ename]) > 0:
                        event_handle = the_map[ename][0]
                        del the_map[ename][0]
                        date = p_title = ""

                        if event_handle:
                            event = self.db.get_event_from_handle(event_handle)
                            date = get_date(event)
                            sortdate = "%09d" % (
                                event.get_date_object().get_sort_value()
                            )
                            place_handle = event.get_place_handle()
                            if place_handle:
                                place = self.db.get_place_from_handle(place_handle)
                                p_title = _pd.display(self.dbstate.db, place)
                        tlist += [date, sortdate, p_title]
                        added = True
                    else:
                        tlist += [""] * 3

                if first:
                    first = False
                    self.row_data.append(tlist)
                elif not added:
                    done = True
                else:
                    self.row_data.append(tlist)
            self.progress_bar.step()

    def make_event_titles(self):
        """
        Create the list of unique event types, along with the person's
        name, birth, and death.
        This should be the column titles of the report.
        """
        the_map = defaultdict(int)
        for individual_id in self.my_list:
            individual = self.db.get_person_from_handle(individual_id)
            for event_ref in individual.get_event_ref_list():
                event = self.db.get_event_from_handle(event_ref.ref)
                name = str(event.get_type())
                if not name:
                    break
                the_map[name] += 1

        unsort_list = sorted(
            [(d, k) for k, d in the_map.items()], key=lambda x: x[0], reverse=True
        )

        sort_list = [item[1] for item in unsort_list]
        ## Presently there's no Birth and Death. Instead there's Birth Date and
        ## Birth Place, as well as Death Date and Death Place.
        ##         # Move birth and death to the begining of the list
        ##         if _("Death") in the_map:
        ##             sort_list.remove(_("Death"))
        ##             sort_list = [_("Death")] + sort_list

        ##         if _("Birth") in the_map:
        ##             sort_list.remove(_("Birth"))
        ##             sort_list = [_("Birth")] + sort_list

        return sort_list

    def on_write_table(self, obj):
        f = Gtk.FileChooserDialog(
            _("Select filename"),
            transient_for=self.window,
            action=Gtk.FileChooserAction.SAVE,
        )
        f.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("_Save"), Gtk.ResponseType.OK
        )

        f.set_current_folder(get_curr_dir())
        status = f.run()
        f.hide()

        if status == Gtk.ResponseType.OK:
            name = f.get_filename()
            doc = ODSTab(len(self.row_data))
            doc.creator(self.db.get_researcher().get_name())

            spreadsheet = TableReport(name, doc)

            new_titles = []
            skip_columns = []
            index = 0
            for title in self.table_titles:
                if title == "sort":
                    skip_columns.append(index)
                else:
                    new_titles.append(title)
                index += 1
            spreadsheet.initialize(len(new_titles))

            spreadsheet.write_table_head(new_titles)

            index = 0
            for top in self.row_data:
                spreadsheet.set_row(index % 2)
                index += 1
                spreadsheet.write_table_data(top, skip_columns)

            spreadsheet.finalize()
        f.destroy()


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class EventComparisonOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "filter": 0,
        }
        filters = utils.get_person_filters(None)
        self.options_help = {
            "filter": (
                "=num",
                "Filter number.",
                [filt.get_name() for filt in filters],
                True,
            ),
        }



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Utilities for getting information from the database.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from ..display.name import displayer as name_displayer
from ..display.place import displayer as place_displayer
from ..lib import EventType, EventRoleType, NameOriginType, Surname

_ = glocale.translation.sgettext

LOG = logging.getLogger(".gui.utils.db")

PRIMARY_EVENT_ROLES = [EventRoleType.FAMILY, EventRoleType.PRIMARY]


# -------------------------------------------------------------------------
#
# Fallback functions
#
# -------------------------------------------------------------------------
def get_birth_or_fallback(db, person, format=None):
    """
    Get BIRTH event from a person, or fallback to an event around
    the time of birth.
    """
    birth_ref = person.get_birth_ref()
    if birth_ref:  # regular birth found
        event = db.get_event_from_handle(birth_ref.ref)
        if event:
            return event

    # now search the event list for fallbacks
    for event_ref in person.get_primary_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event and event.type.is_birth_fallback() and event_ref.role.is_primary():
            if format:
                event.date.format = format
            return event
    return None


def get_death_or_fallback(db, person, format=None):
    """
    Get a DEATH event from a person, or fallback to an
    event around the time of death.
    """
    death_ref = person.get_death_ref()
    if death_ref:  # regular death found
        event = db.get_event_from_handle(death_ref.ref)
        if event:
            return event

    # now search the event list for fallbacks
    for event_ref in person.get_primary_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event and event.type.is_death_fallback() and event_ref.role.is_primary():
            if format:
                event.date.format = format
            return event
    return None


def get_age(db, person, fallback=True, calendar="gregorian"):
    """
    Compute the age of person.

    :param person: person handle or person object
    :param fallback: Allow fallback events if True
    :returns: tuple of year, month day if valid, None otherwise
    """
    birth = None
    death = None
    if isinstance(person, str):
        # a handle is passed
        person = db.get_person_from_handle(person)
    if fallback:
        birth = get_birth_or_fallback(db, person)
        death = get_death_or_fallback(db, person)
    else:
        birth_ref = person.get_birth_ref()
        if birth_ref:  # regular birth found
            birth = db.get_event_from_handle(birth_ref.ref)
        death_ref = person.get_death_ref()
        if death_ref:  # regular death found
            death = db.get_event_from_handle(death_ref.ref)

    age = None
    if birth is not None:
        birth_date = birth.get_date_object().to_calendar("gregorian")
        if birth_date and birth_date.get_valid() and not birth_date.is_empty():
            if death is not None:
                death_date = death.get_date_object().to_calendar("gregorian")
                if death_date and death_date.get_valid() and not death_date.is_empty():
                    age = death_date - birth_date
                    if not age.is_valid():
                        age = None
                    else:
                        age = age.tuple()
    return age


def get_timeperiod(db, person):
    """
    Compute the timeperiod a person lived in

    :param person: person handle or person object
    :returns: the year, None otherwise
    """
    if isinstance(person, str):
        # a handle is passed
        person = db.get_person_from_handle(person)

    # the period is the year of birth
    birth = get_birth_or_fallback(db, person)
    if birth is not None:
        birth_date = birth.get_date_object().to_calendar("gregorian")
        if birth_date and birth_date.get_valid() and not birth_date.is_empty():
            return birth_date.get_year()

    death = get_death_or_fallback(db, person)
    # no birth, period is death - 20
    if death is not None:
        death_date = death.get_date_object().to_calendar("gregorian")
        if death_date and death_date.get_valid() and not death_date.is_empty():
            return death_date.get_year() - 20

    # no birth and death, look for another event date we can use
    for event_ref in person.get_primary_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event:
            event_date = event.get_date_object().to_calendar("gregorian")
            if event_date and event_date.get_valid() and not event_date.is_empty():
                return event_date.get_year()
    return None


def get_event_ref(db, family, event_type):
    """
    Return a reference to a primary family event of the given event type.
    """
    if family:
        for event_ref in family.get_event_ref_list():
            event = db.get_event_from_handle(event_ref.ref)
            if (
                event
                and event.get_type() == event_type
                and event_ref.get_role() in PRIMARY_EVENT_ROLES
            ):
                return event_ref
    return None


def get_primary_event_ref_list(db, family):
    """
    Return a reference to the primary events of the family.
    """
    events = []
    for event_ref in family.get_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event and event_ref.get_role() in PRIMARY_EVENT_ROLES:
            events.append(event_ref)
    return events


def get_marriage_or_fallback(db, family, format=None):
    """
    Get a MARRIAGE event from a family, or fallback to an
    alternative event type.
    """
    marriage_ref = get_event_ref(db, family, EventType.MARRIAGE)
    if marriage_ref:  # regular marriage found
        event = db.get_event_from_handle(marriage_ref.ref)
        if event:
            return event

    # now search the event list for fallbacks
    for event_ref in get_primary_event_ref_list(db, family):
        event = db.get_event_from_handle(event_ref.ref)
        if (
            event
            and event.type.is_marriage_fallback()
            and event_ref.role in PRIMARY_EVENT_ROLES
        ):
            if format:
                event.date.format = format
            return event
    return None


def get_divorce_or_fallback(db, family, format=None):
    """
    Get a DIVORCE event from a family, or fallback to an
    alternative event type.
    """
    divorce_ref = get_event_ref(db, family, EventType.DIVORCE)
    if divorce_ref:  # regular marriage found
        event = db.get_event_from_handle(divorce_ref.ref)
        if event:
            return event

    # now search the event list for fallbacks
    for event_ref in get_primary_event_ref_list(db, family):
        event = db.get_event_from_handle(event_ref.ref)
        if (
            event
            and event.type.is_divorce_fallback()
            and event_ref.role in PRIMARY_EVENT_ROLES
        ):
            if format:
                event.date.format = format
            return event
    return None


# -------------------------------------------------------------------------
#
# Function to return the name of the main participant of an event
#
# -------------------------------------------------------------------------
def get_participant_from_event(db, event_handle, all_=False):
    """
    Obtain the first primary or family participant to an event we find in the
    database. Note that an event can have more than one primary or
    family participant, only one is returned, adding ellipses if there are
    more. If the all_ parameter is true a comma-space separated string with
    the names of all primary participants is returned and no ellipses is used.
    """
    participant = ""
    ellipses = False
    result_list = list(
        db.find_backlink_handles(event_handle, include_classes=["Person", "Family"])
    )

    # obtain handles without duplicates
    people = set([x[1] for x in result_list if x[0] == "Person"])
    families = set([x[1] for x in result_list if x[0] == "Family"])
    for person_handle in people:
        person = db.get_person_from_handle(person_handle)
        if not person:
            continue
        for event_ref in person.get_event_ref_list():
            if event_handle == event_ref.ref and event_ref.get_role().is_primary():
                if participant:
                    if all_:
                        participant += f", {name_displayer.display(person)}"
                    else:
                        ellipses = True
                else:
                    participant = name_displayer.display(person)
                break
        if ellipses:
            break
    if ellipses:
        return _("%s, ...") % participant

    for family_handle in families:
        family = db.get_family_from_handle(family_handle)
        for event_ref in family.get_event_ref_list():
            if event_handle == event_ref.ref and event_ref.get_role().is_family():
                if participant:
                    if all_:
                        participant += f", {family_name(family, db)}"
                    else:
                        ellipses = True
                else:
                    participant = family_name(family, db)
                break
        if ellipses:
            break

    if ellipses:
        return _("%s, ...") % participant
    return participant


# -------------------------------------------------------------------------
#
# Function to return a label to display the active object in the status bar
# and to describe bookmarked objects.
#
# -------------------------------------------------------------------------
def navigation_label(db, nav_type, handle_or_obj):
    """
    Return a descriptive label for an object.
    """
    if isinstance(handle_or_obj, str):
        get_method = db.method("get_%s_from_handle", nav_type)
        obj = get_method(handle_or_obj)
    else:
        obj = handle_or_obj

    if obj is None:
        return (None, None)

    label = None
    if nav_type == "Person":
        label = name_displayer.display(obj)
    elif nav_type == "Family":
        label = family_name(obj, db)
    elif nav_type == "Event":
        who = get_participant_from_event(db, obj.get_handle())
        desc = obj.get_description()
        label = obj.get_type()
        if desc:
            label = f"{label} - {desc}"
        if who:
            label = f"{label} - {who}"
    elif nav_type == "Place":
        label = place_displayer.display(db, obj)
    elif nav_type == "Source":
        label = obj.get_title()
    elif nav_type == "Citation":
        label = obj.get_page()
        src = db.get_source_from_handle(obj.get_reference_handle())
        if src:
            label = f"{src.get_title()} {label}"
    elif nav_type == "Repository":
        label = obj.get_name()
    elif nav_type == "Media":
        label = obj.get_description()
    elif nav_type == "Note":
        label = obj.get()
        # When strings are cut, make sure they are unicode
        # otherwise you may end of with cutting within an utf-8 sequence
        label = str(label)
        label = " ".join(label.split())
        if len(label) > 40:
            label = label[:40] + "..."

    if label and obj:
        label = f"[{obj.get_gramps_id()}] {label}"

    return (label, obj)


# -------------------------------------------------------------------------
#
# Function to return children's list of a person
#
# -------------------------------------------------------------------------
def find_children(db, person):
    """
    Return the list of all children's IDs for a person.
    """
    children = set()
    for family_handle in person.get_family_handle_list():
        family = db.get_family_from_handle(family_handle)
        if family:
            for child_ref in family.get_child_ref_list():
                children.add(child_ref.ref)
    return list(children)


# -------------------------------------------------------------------------
#
# Function to return parent's list of a person
#
# -------------------------------------------------------------------------
def find_parents(db, person):
    """
    Return the unique list of all parents' IDs for a person.
    """
    parents = set()
    for family_handle in person.get_parent_family_handle_list():
        family = db.get_family_from_handle(family_handle)
        father_handle = family.get_father_handle()
        if father_handle:
            parents.add(father_handle)
        mother_handle = family.get_mother_handle()
        if mother_handle:
            parents.add(mother_handle)
    return list(parents)


# -------------------------------------------------------------------------
#
# Function to return persons, that share the same event.
# This for example links witnesses to the tree
#
# -------------------------------------------------------------------------
def find_witnessed_people(db, person):
    """
    Return list of all person handles associated with an event.
    """
    person_handle = person.get_handle()
    people = {
        person_ref.ref
        for person_ref in person.get_person_ref_list()
        if person_ref.ref != person_handle
    }
    for event_ref in person.get_event_ref_list():
        for backlink in db.find_backlink_handles(
            event_ref.ref, include_classes=["Person", "Family"]
        ):
            if backlink[0] == "Person":
                if backlink[1] != person_handle:
                    people.add(backlink[1])
                continue
            if backlink[0] == "Family":
                family = db.get_family_from_handle(backlink[1])
                if family:
                    father_handle = family.get_father_handle()
                    if father_handle and father_handle != person_handle:
                        people.add(father_handle)
                    mother_handle = family.get_mother_handle()
                    if mother_handle and mother_handle != person_handle:
                        people.add(mother_handle)

    for family_handle in person.get_family_handle_list():
        family = db.get_family_from_handle(family_handle)
        for event_ref in family.get_event_ref_list():
            for backlink in db.find_backlink_handles(
                event_ref.ref, include_classes=["Person"]
            ):
                if backlink[1] != person_handle:
                    people.add(backlink[1])

    return list(people)


# -------------------------------------------------------------------------
#
#  Iterate over ancestors.
#
# -------------------------------------------------------------------------
def for_each_ancestor(db, start, func, data):
    """
    Recursively iterate (breadth-first) over ancestors of
    people listed in start.
    Call func(data, pid) for the Id of each person encountered.
    Exit and return 1, as soon as func returns true.
    Return 0 otherwise.
    """
    todo = start
    done_ids = set()
    while len(todo):
        person_handle = todo.pop()
        # Don't process the same handle twice.  This can happen
        # if there is a cycle in the database, or if the
        # initial list contains X and some of X's ancestors.
        if person_handle in done_ids:
            continue
        if func(data, person_handle):
            return 1
        person = db.get_person_from_handle(person_handle)
        for family_handle in person.get_parent_family_handle_list():
            family = db.get_family_from_handle(family_handle)
            if family:
                father_handle = family.get_father_handle()
                if father_handle:
                    todo.append(father_handle)
                mother_handle = family.get_mother_handle()
                if mother_handle:
                    todo.append(mother_handle)
        done_ids.add(person_handle)
    return 0


# -------------------------------------------------------------------------
#
# Preset a name with a name of family member
#
# -------------------------------------------------------------------------
def preset_name(base_person, name, sibling=False):
    """
    Fill up name with all family common names of base person.
    If sibling=True, pa/matronymics are retained.
    """
    surnames = []
    primary_name = base_person.get_primary_name()
    primary = False
    for surname in primary_name.get_surname_list():
        if (not sibling) and (
            surname.get_origintype().value
            in [NameOriginType.PATRONYMIC, NameOriginType.MATRONYMIC]
        ):
            continue
        surnames.append(Surname(source=surname))
        if surname.primary:
            primary = True
    if not surnames:
        surnames = [Surname()]
    name.set_surname_list(surnames)
    if not primary:
        name.set_primary_surname(0)
    name.set_family_nick_name(primary_name.get_family_nick_name())
    name.set_group_as(primary_name.get_group_as())
    name.set_sort_as(primary_name.get_sort_as())


# -------------------------------------------------------------------------
#
# Short hand function to return either the person's name, or an empty
# string if the person is None
#
# -------------------------------------------------------------------------
def family_name(family, db, noname=_("unknown")):
    """
    Builds a name for the family from the parents names.
    """
    father = None
    father_handle = family.get_father_handle()
    if father_handle:
        father = db.get_person_from_handle(father_handle)

    mother = None
    mother_handle = family.get_mother_handle()
    if mother_handle:
        mother = db.get_person_from_handle(mother_handle)

    if father and mother:
        fname = name_displayer.display(father)
        mname = name_displayer.display(mother)
        return _("%(father)s and %(mother)s") % {
            "father": fname,
            "mother": mname,
        }
    if father:
        return name_displayer.display(father)
    if mother:
        return name_displayer.display(mother)
    return noname


# -------------------------------------------------------------------------
#
# Referents functions
#
# -------------------------------------------------------------------------
def get_referents(handle, db, primary_objects):
    """
    Find objects that refer to an object.

    This function is the base for other get_<object>_referents functions.
    """
    # Use one pass through the reference map to grab all the references
    object_list = list(db.find_backlink_handles(handle))

    # Then form the object-specific lists
    the_lists = ()

    for primary in primary_objects:
        primary_list = [item[1] for item in object_list if item[0] == primary]
        the_lists = the_lists + (primary_list,)
    return the_lists


def get_source_referents(source_handle, db):
    """
    Find objects that refer the source.

    This function finds all primary objects that refer (directly or through
    secondary child-objects) to a given source handle in a given database.

    Only Citations can refer to sources, so that is all we need to check
    """
    _primaries = ("Citation",)

    return get_referents(source_handle, db, _primaries)


def get_citation_referents(citation_handle, db):
    """
    Find objects that refer the citation.

    This function finds all primary objects that refer (directly or through
    secondary child-objects) to a given citation handle in a given database.
    """
    _primaries = (
        "Person",
        "Family",
        "Event",
        "Place",
        "Source",
        "Media",
        "Repository",
    )

    return get_referents(citation_handle, db, _primaries)


def get_source_and_citation_referents(source_handle, db):
    """
    Find all citations that refer to the sources, and recursively, all objects
    that refer to the sources.

    This function finds all primary objects that refer (directly or through
    secondary child-objects) to a given source handle in a given database.

    | Objects -> Citations -> Source
    | e.g.
    | Media object M1  -> Citation C1 -> Source S1
    | Media object M2  -> Citation C1 -> Source S1
    | Person object P1 -> Citation C2 -> Source S1

    The returned structure is rather ugly, but provides all the information in
    a way that is consistent with the other Util functions.

    | (
    | tuple of objects that refer to the source - only first element is present
    |     ([C1, C2],),
    | list of citations with objects that refer to them
    |     [
    |         (C1,
    |             tuple of reference lists
    |               P,  F,  E,  Pl, S,  M,        R
    |             ([], [], [], [], [], [M1, M2]. [])
    |         )
    |         (C2,
    |             tuple of reference lists
    |               P,    F,  E,  Pl, S,  M,  R
    |             ([P1], [], [], [], [], []. [])
    |         )
    |     ]
    | )
    """
    the_lists = get_source_referents(source_handle, db)
    LOG.debug("source referents %s", [the_lists])
    # now, for each citation, get the objects that refer to that citation
    citation_referents_list = []
    for citation in the_lists[0]:
        LOG.debug("citation %s", citation)
        refs = get_citation_referents(citation, db)
        citation_referents_list += [(citation, refs)]
    LOG.debug("citation_referents_list %s", [citation_referents_list])

    (citation_list) = the_lists
    the_lists = (citation_list, citation_referents_list)

    LOG.debug("the_lists %s", [the_lists])
    return the_lists


def get_media_referents(media_handle, db):
    """
    Find objects that refer the media object.

    This function finds all primary objects that refer
    to a given media handle in a given database.
    """
    _primaries = ("Person", "Family", "Event", "Place", "Source", "Citation")

    return get_referents(media_handle, db, _primaries)


def get_note_referents(note_handle, db):
    """
    Find objects that refer a note object.

    This function finds all primary objects that refer
    to a given note handle in a given database.
    """
    _primaries = (
        "Person",
        "Family",
        "Event",
        "Place",
        "Source",
        "Citation",
        "Media",
        "Repository",
    )

    return get_referents(note_handle, db, _primaries)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006 Donald N. Allingham
# Copyright (C) 2007-2012 Brian G. Matherly
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2012-2014 Paul Franklin
# Copyright (C) 2012      Nicolas Adenis-Lamarre
# Copyright (C) 2012      Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

""" fanchart report """

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from math import pi, cos, sin, log10, acos, asin, degrees, radians


def log2(val):
    """
    Calculate the log base 2 of a value.
    """
    return int(log10(val) / log10(2))


# ------------------------------------------------------------------------
#
# gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import ReportError
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    GraphicsStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
    IndexMark,
    INDEX_TYPE_TOC,
)
from gramps.gen.plug.menu import (
    EnumeratedListOption,
    NumberOption,
    PersonOption,
    BooleanOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.config import config
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.lib import EventType, Date
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.display.name import displayer as _nd

# ------------------------------------------------------------------------
#
# private constants
#
# ------------------------------------------------------------------------
FULL_CIRCLE = 0
HALF_CIRCLE = 1
QUAR_CIRCLE = 2
OVERHANG = 3

BACKGROUND_WHITE = 0
BACKGROUND_GEN = 1

RADIAL_UPRIGHT = 0
RADIAL_ROUNDABOUT = 1

# minor offset just usefull for generation 11,
# to not a bit offset between the text and the polygon
# this can be considered as a bad hack
WEDGE_TEXT_BARRE_OFFSET = 0.0016


# ------------------------------------------------------------------------
#
# private functions
#
# ------------------------------------------------------------------------
def draw_wedge(
    doc,
    style,
    centerx,
    centery,
    radius,
    start_angle,
    end_angle,
    do_rendering,
    short_radius=0,
):
    """
    Draw a wedge shape.
    """
    while end_angle < start_angle:
        end_angle += 360

    path = []

    radiansdelta = radians(0.5)
    sangle = radians(start_angle)
    eangle = radians(end_angle)
    while eangle < sangle:
        eangle = eangle + 2 * pi
    angle = sangle

    if short_radius == 0:
        if (end_angle - start_angle) != 360:
            path.append((centerx, centery))
    else:
        origx = centerx + cos(angle) * short_radius
        origy = centery + sin(angle) * short_radius
        path.append((origx, origy))

    while angle < eangle:
        _x_ = centerx + cos(angle) * radius
        _y_ = centery + sin(angle) * radius
        path.append((_x_, _y_))
        angle = angle + radiansdelta
    _x_ = centerx + cos(eangle) * radius
    _y_ = centery + sin(eangle) * radius
    path.append((_x_, _y_))

    if short_radius:
        _x_ = centerx + cos(eangle) * short_radius
        _y_ = centery + sin(eangle) * short_radius
        path.append((_x_, _y_))

        angle = eangle
        while angle >= sangle:
            _x_ = centerx + cos(angle) * short_radius
            _y_ = centery + sin(angle) * short_radius
            path.append((_x_, _y_))
            angle -= radiansdelta
    if do_rendering:
        doc.draw_path(style, path)

    delta = (eangle - sangle) / 2.0
    rad = short_radius + (radius - short_radius) / 2.0

    return (
        (centerx + cos(sangle + delta + WEDGE_TEXT_BARRE_OFFSET) * rad),
        (centery + sin(sangle + delta + WEDGE_TEXT_BARRE_OFFSET) * rad),
    )


# ------------------------------------------------------------------------
#
# FanChart
#
# ------------------------------------------------------------------------
class FanChart(Report):
    def __init__(self, database, options, user):
        """
        Create the FanChart object that produces the report.

        The arguments are:

        database     - the Gramps database instance
        options      - instance of the Options class for this report
        user         - a gen.user.User instance

        This report needs the following parameters (class variables)
        that come in the options class.

        maxgen       - Maximum number of generations to include.
        circle       - Draw a full circle, half circle, half circle with
                       overhang or quarter circle.
        background   - Background color is generation dependent or white.
        radial       - Print radial texts roundabout or as upright as possible.
        draw_empty   - draw background when there is no information
        same_style   - use the same style for all generation
        flip_text    - flip names that are on the bottom half of the fan
        incl_private - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """

        Report.__init__(self, database, options, user)

        menu = options.menu

        self.set_locale(options.menu.get_option_by_name("trans").get_value())

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)

        self.max_generations = menu.get_option_by_name("maxgen").get_value()
        self.circle = menu.get_option_by_name("circle").get_value()
        self.background = menu.get_option_by_name("background").get_value()
        self.radial = menu.get_option_by_name("radial").get_value()
        pid = menu.get_option_by_name("pid").get_value()
        self.draw_empty = menu.get_option_by_name("draw_empty").get_value()
        self.same_style = menu.get_option_by_name("same_style").get_value()
        self.flip_text = menu.get_option_by_name("flip_text").get_value()
        self.exclude_title = menu.get_option_by_name("exclude_title").get_value()
        self.center_person = self.database.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        self.graphic_style = []
        self.text_style = []
        for i in range(0, self.max_generations):
            self.graphic_style.append("FC-Graphic" + "%02d" % i)
            self.text_style.append("FC-Text" + "%02d" % i)

        self.calendar = 0

        self.height = 0
        self.map = [None] * 2**self.max_generations
        self.text = {}

    def apply_filter(self, person_handle, index):
        """traverse the ancestors recursively until either the end
        of a line is found, or until we reach the maximum number of
        generations that we want to deal with"""

        if (not person_handle) or (index >= 2**self.max_generations):
            return
        self.map[index - 1] = person_handle
        self.text[index - 1] = self.get_info(person_handle, log2(index))

        person = self.database.get_person_from_handle(person_handle)
        family_handle = person.get_main_parents_family_handle()
        if family_handle:
            family = self.database.get_family_from_handle(family_handle)
            self.apply_filter(family.get_father_handle(), index * 2)
            self.apply_filter(family.get_mother_handle(), (index * 2) + 1)

    def write_report(self):
        self.doc.start_page()

        self.apply_filter(self.center_person.get_handle(), 1)
        p_rn = self.center_person.get_primary_name().get_regular_name()

        chart_height = self.doc.get_usable_height()
        if self.exclude_title:
            title = ""
            title_nb_lines = 0
            title_height = 0
        else:
            # choose one line or two lines translation according to the width
            title = self._("%(generations)d Generation Fan Chart " "for %(person)s") % {
                "generations": self.max_generations,
                "person": p_rn,
            }
            title_nb_lines = 1
            title_height = 0
            style_sheet = self.doc.get_style_sheet()
            if style_sheet:
                p_style = style_sheet.get_paragraph_style("FC-Title")
                if p_style:
                    font = p_style.get_font()
                    if font:
                        title_width = utils.pt2cm(self.doc.string_width(font, title))
                        if title_width > self.doc.get_usable_width():
                            title = self._(
                                "%(generations)d Generation Fan Chart "
                                "for\n%(person)s"
                            ) % {"generations": self.max_generations, "person": p_rn}
                            title_nb_lines = 2
                        fontsize = utils.pt2cm(font.get_size())
                        title_height = fontsize * title_nb_lines * 1.5
            chart_height -= title_height

        if self.circle == FULL_CIRCLE:
            max_angle = 360.0
            start_angle = 90
            max_circular = 5
            _x_ = self.doc.get_usable_width() / 2.0
            _y_ = chart_height / 2.0
            min_xy = min(_x_, _y_)

        elif self.circle == HALF_CIRCLE:
            max_angle = 180.0
            start_angle = 180
            max_circular = 3
            _x_ = self.doc.get_usable_width() / 2.0
            _y_ = chart_height
            min_xy = min(_x_, _y_)

        elif self.circle == OVERHANG:
            if (
                chart_height < self.doc.get_usable_width()
                and chart_height > self.doc.get_usable_width() / 2
            ):
                # Determine overhang angle to fill the paper
                radius = self.doc.get_usable_width() / 2
                _overhang_height_ = chart_height - radius
                overhang_angle = degrees(asin(_overhang_height_ / radius))
            else:
                overhang_angle = 25.0
            max_angle = 180 + 2 * overhang_angle
            start_angle = 180 - overhang_angle
            max_circular = 3
            _x_ = self.doc.get_usable_width() / 2.0
            _overhang_height_ = sin(radians(overhang_angle)) * _x_
            _y_ = chart_height - _overhang_height_
            min_xy = min(_x_, _y_)

        else:  # quarter circle
            max_angle = 90.0
            start_angle = 270
            max_circular = 2
            _x_ = 0
            _y_ = chart_height
            min_xy = min(self.doc.get_usable_width(), _y_)

        if self.max_generations > max_circular:
            block_size = min_xy / (self.max_generations * 2 - max_circular)
        else:
            block_size = min_xy / self.max_generations

        # adaptation of the fonts (title and others)
        optimized_style_sheet = self.get_optimized_style_sheet(
            title,
            max_circular,
            block_size,
            self.same_style,
            not self.same_style,
            # if same_style, use default generated colors
            self.background == BACKGROUND_WHITE,
        )

        if optimized_style_sheet:
            self.doc.set_style_sheet(optimized_style_sheet)

        # title
        if title_height:
            mark = IndexMark(title, INDEX_TYPE_TOC, 1)
            titlex = self.doc.get_usable_width() / 2
            titley = 0
            self.doc.center_text("FC-Graphic-title", title, titlex, titley, mark)
            _y_ += title_height

        # wheel
        for generation in range(0, min(max_circular, self.max_generations)):
            self.draw_circular(_x_, _y_, start_angle, max_angle, block_size, generation)
        for generation in range(max_circular, self.max_generations):
            self.draw_radial(_x_, _y_, start_angle, max_angle, block_size, generation)
        self.doc.end_page()

    def get_info(self, person_handle, generation):
        """get info about a person"""
        person = self.database.get_person_from_handle(person_handle)
        p_pn = person.get_primary_name()
        self.calendar = config.get("preferences.calendar-format-report")

        birth = get_birth_or_fallback(self.database, person)
        bth = ""
        if birth:
            bth = birth.get_date_object()
            bth = self._get_date(
                Date(bth.to_calendar(self.calendar).get_year())
            )  # localized
            if bth == 0:
                bth = ""
            elif birth.get_type() != EventType.BIRTH:
                bth += "*"

        death = get_death_or_fallback(self.database, person)
        dth = ""
        if death:
            dth = death.get_date_object()
            dth = self._get_date(
                Date(dth.to_calendar(self.calendar).get_year())
            )  # localized
            if dth == 0:
                dth = ""
            elif death.get_type() != EventType.DEATH:
                dth += "*"
        if bth and dth:
            val = "%s - %s" % (str(bth), str(dth))
        elif bth:
            val = "* %s" % (str(bth))
        elif dth:
            val = "+ %s" % (str(dth))
        else:
            val = ""

        if generation > 7:
            if (p_pn.get_first_name() != "") and (p_pn.get_surname() != ""):
                name = p_pn.get_first_name() + " " + p_pn.get_surname()
            else:
                name = p_pn.get_first_name() + p_pn.get_surname()
            if (name != "") and (val != ""):
                string = name + self._(", ") + val  # Arabic OK
            else:
                string = name + val
            return [string]
        elif generation == 7:
            if (p_pn.get_first_name() != "") and (p_pn.get_surname() != ""):
                name = p_pn.get_first_name() + " " + p_pn.get_surname()
            else:
                name = p_pn.get_first_name() + p_pn.get_surname()

            if self.circle == FULL_CIRCLE:
                return [name, val]
            elif self.circle in (HALF_CIRCLE, OVERHANG):
                return [name, val]
            else:
                if (name != "") and (val != ""):
                    string = name + self._(", ") + val  # Arabic OK
                else:
                    string = name + val
                return [string]
        elif generation == 6:
            if self.circle == FULL_CIRCLE:
                return [p_pn.get_first_name(), p_pn.get_surname(), val]
            elif self.circle in (HALF_CIRCLE, OVERHANG):
                return [p_pn.get_first_name(), p_pn.get_surname(), val]
            else:
                if (p_pn.get_first_name() != "") and (p_pn.get_surname() != ""):
                    name = p_pn.get_first_name() + " " + p_pn.get_surname()
                else:
                    name = p_pn.get_first_name() + p_pn.get_surname()
                return [name, val]
        else:
            return [p_pn.get_first_name(), p_pn.get_surname(), val]

    def get_max_width_for_circles(self, rad1, rad2, max_centering_proportion):
        r"""
        (the "r" in the above line is to keep pylint happy)
           __
          /__\ <- compute the line width which is drawable between 2 circles.
         /  _ \   max_centering_proportion : 0, touching the circle1, 1,
        |  |_| |   touching the circle2, 0.5 : middle between the 2 circles
        |      |
         \    /
          \__/
                 basically, max_centering_proportion is
                 max_centering_proportion/nb_lines
        """
        # radius at the center of the 2 circles
        rmid = rad2 - (rad2 - rad1) * max_centering_proportion
        return sin(acos(rmid / rad2)) * rad2 * 2

    def get_max_width_for_circles_line(
        self, rad1, rad2, line, nb_lines, centering=False
    ):
        r"""
        (the "r" in the above line is to keep pylint happy)
           __
          /__\ <- compute the line width which is drawable between 2 circles.
         /  _ \   instead of a max_centering_proportion, you get a
        |  |_| |  line/nb_lines position.  (we suppose that lines have the
        |      |  same heights.) for example, if you've 2 lines to draw,
         \    /   line 2 max width is at the 2/3 between the 2 circles
          \__/
        """
        if centering:
            return self.get_max_width_for_circles(rad1, rad2, 1.0)
        else:
            return self.get_max_width_for_circles(
                rad1, rad2, line / float(nb_lines + 1)
            )

    def get_optimized_font_size_for_text(self, rad1, rad2, text, font, centering=False):
        """
        a text can be several lines
        find the font size equals or lower than font.get_size() which fit
        between rad1 and rad2 to display the text.
        centering is a special case when you've the full circle
        available to draw the text in the middle of it
        """
        min_font_size = font.get_size()
        i = 1
        nb_lines = len(text)
        for line in text:
            font_size = self.get_optimized_font_size(
                line,
                font,
                self.get_max_width_for_circles_line(rad1, rad2, i, nb_lines, centering),
            )
            i += 1
            if min_font_size > font_size:
                min_font_size = font_size
        return min_font_size

    def get_optimized_font_size(self, line, font, max_width):
        """
        for a given width, guess the best font size which is equals
        or smaller than font which make line fit into max_width
        """
        test_font = FontStyle(font)
        width = utils.pt2cm(self.doc.string_width(test_font, line))
        while width > max_width and test_font.get_size() > 1:
            test_font.set_size(test_font.get_size() - 1)
            width = utils.pt2cm(self.doc.string_width(test_font, line))
        return test_font.get_size()

    def get_optimized_style_sheet(
        self,
        title,
        max_circular,
        block_size,
        map_style_from_single,
        map_paragraphs_colors_to_graphics,
        make_background_white,
    ):
        """
        returns an optimized (modified) style sheet which make fanchart
        look nicer
        """
        new_style_sheet = self.doc.get_style_sheet()
        if not new_style_sheet:
            return self.doc.get_style_sheet()

        # update title font size
        pstyle_name = "FC-Title"
        p_style = new_style_sheet.get_paragraph_style(pstyle_name)
        if p_style:
            title_font = p_style.get_font()
            if title_font:
                title_width = utils.pt2cm(
                    self.doc.string_multiline_width(title_font, title)
                )
                while (
                    title_width > self.doc.get_usable_width()
                    and title_font.get_size() > 1
                ):
                    title_font.set_size(title_font.get_size() - 1)
                    title_width = utils.pt2cm(
                        self.doc.string_multiline_width(title_font, title)
                    )
                new_style_sheet.add_paragraph_style(pstyle_name, p_style)

        # biggest font allowed is the one of the fist generation, after,
        # always lower than the previous one
        p_style = new_style_sheet.get_paragraph_style(self.text_style[0])
        font = None
        if p_style:
            font = p_style.get_font()
        if font:
            previous_generation_font_size = font.get_size()

            for generation in range(0, self.max_generations):
                gstyle_name = self.graphic_style[generation]
                pstyle_name = self.text_style[generation]
                g_style = new_style_sheet.get_draw_style(gstyle_name)

                # p_style is a copy of 'FC-Text' - use different style
                # to be able to auto change some fonts for some generations
                if map_style_from_single:
                    p_style = new_style_sheet.get_paragraph_style("FC-Text")
                else:
                    p_style = new_style_sheet.get_paragraph_style(pstyle_name)

                if g_style and p_style:
                    # set graphic colors to paragraph colors,
                    # while it's functionnaly
                    # the same for fanchart or make backgrounds white
                    if make_background_white:
                        g_style.set_fill_color((255, 255, 255))
                        new_style_sheet.add_draw_style(gstyle_name, g_style)
                    elif map_paragraphs_colors_to_graphics:
                        pstyle = new_style_sheet.get_paragraph_style(pstyle_name)
                        if pstyle:
                            g_style.set_fill_color(pstyle.get_background_color())
                            new_style_sheet.add_draw_style(gstyle_name, g_style)

                    # adapt font size if too big
                    segments = 2**generation
                    if generation < min(max_circular, self.max_generations):
                        # adpatation for circular fonts
                        rad1, rad2 = self.get_circular_radius(
                            block_size, generation, self.circle
                        )
                        font = p_style.get_font()
                        if font:
                            min_font_size = font.get_size()
                            # find the smallest font required
                            for index in range(segments - 1, 2 * segments - 1):
                                if self.map[index]:
                                    font_size = self.get_optimized_font_size_for_text(
                                        rad1,
                                        rad2,
                                        self.text[index],
                                        p_style.get_font(),
                                        (
                                            self.circle == FULL_CIRCLE
                                            and generation == 0
                                        ),
                                    )
                                if font_size < min_font_size:
                                    min_font_size = font_size
                            font.set_size(
                                min(previous_generation_font_size, min_font_size)
                            )
                    else:
                        # adaptation for radial fonts

                        # find the largest string for the generation
                        longest_line = ""
                        longest_width = 0
                        for index in range(segments - 1, 2 * segments - 1):
                            if self.map[index]:
                                for line in self.text[index]:
                                    width = utils.pt2cm(
                                        self.doc.string_multiline_width(
                                            p_style.get_font(), line
                                        )
                                    )
                                    if width > longest_width:
                                        longest_line = line
                                        longest_width = width

                        # determine maximum width allowed for this generation
                        rad1, rad2 = self.get_radial_radius(
                            block_size, generation, self.circle
                        )
                        max_width = rad2 - rad1

                        # reduce the font so that longest_width
                        # fit into max_width
                        font = p_style.get_font()
                        if font:
                            font.set_size(
                                min(
                                    previous_generation_font_size,
                                    self.get_optimized_font_size(
                                        longest_line, p_style.get_font(), max_width
                                    ),
                                )
                            )

                    # redefine the style
                    new_style_sheet.add_paragraph_style(pstyle_name, p_style)
                    font = p_style.get_font()
        if font:
            previous_generation_font_size = font.get_size()

        # finished
        return new_style_sheet

    def draw_circular(self, _x_, _y_, start_angle, max_angle, size, generation):
        segments = 2**generation
        delta = max_angle / segments
        end_angle = start_angle
        text_angle = start_angle - 270 + (delta / 2.0)
        rad1, rad2 = self.get_circular_radius(size, generation, self.circle)
        graphic_style = self.graphic_style[generation]

        for index in range(segments - 1, 2 * segments - 1):
            start_angle = end_angle
            end_angle = start_angle + delta
            (_xc, _yc) = draw_wedge(
                self.doc,
                graphic_style,
                _x_,
                _y_,
                rad2,
                start_angle,
                end_angle,
                self.map[index] or self.draw_empty,
                rad1,
            )
            if self.map[index]:
                if (generation == 0) and self.circle == FULL_CIRCLE:
                    _yc = _y_
                person = self.database.get_person_from_handle(self.map[index])
                mark = utils.get_person_mark(self.database, person)

                if (
                    self.flip_text == True
                    and generation >= 2
                    and (
                        start_angle >= 360 or (start_angle >= 90 and start_angle < 180)
                    )
                ):
                    self.doc.rotate_text(
                        graphic_style,
                        self.text[index],
                        _xc,
                        _yc,
                        text_angle + 180,
                        mark,
                    )
                else:
                    self.doc.rotate_text(
                        graphic_style, self.text[index], _xc, _yc, text_angle, mark
                    )
            text_angle += delta

    def get_radial_radius(self, size, generation, circle):
        """determine the radius"""
        if circle == FULL_CIRCLE:
            rad1 = size * ((generation * 2) - 5)
            rad2 = size * ((generation * 2) - 3)
        elif circle in (HALF_CIRCLE, OVERHANG):
            rad1 = size * ((generation * 2) - 3)
            rad2 = size * ((generation * 2) - 1)
        else:  # quarter circle
            rad1 = size * ((generation * 2) - 2)
            rad2 = size * (generation * 2)
        return rad1, rad2

    def get_circular_radius(self, size, generation, circle):
        """determine the radius"""
        return size * generation, size * (generation + 1)

    def draw_radial(self, _x_, _y_, start_angle, max_angle, size, generation):
        segments = 2**generation
        delta = max_angle / segments
        end_angle = start_angle
        text_angle = start_angle - delta / 2.0
        graphic_style = self.graphic_style[generation]

        rad1, rad2 = self.get_radial_radius(size, generation, self.circle)
        for index in range(segments - 1, 2 * segments - 1):
            start_angle = end_angle
            end_angle = start_angle + delta
            (_xc, _yc) = draw_wedge(
                self.doc,
                graphic_style,
                _x_,
                _y_,
                rad2,
                start_angle,
                end_angle,
                self.map[index] or self.draw_empty,
                rad1,
            )
            text_angle += delta
            if self.map[index]:
                person = self.database.get_person_from_handle(self.map[index])
                mark = utils.get_person_mark(self.database, person)
                if (
                    self.radial == RADIAL_UPRIGHT
                    and (start_angle >= 90)
                    and (start_angle < 270)
                ):
                    self.doc.rotate_text(
                        graphic_style,
                        self.text[index],
                        _xc,
                        _yc,
                        text_angle + 180,
                        mark,
                    )
                else:
                    self.doc.rotate_text(
                        graphic_style, self.text[index], _xc, _yc, text_angle, mark
                    )


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class FanChartOptions(MenuReportOptions):
    """options for fanchart report"""

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.max_generations = 11
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        """
        Add options to the menu for the fan chart.
        """
        category_name = _("Report Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)

        max_gen = NumberOption(_("Generations"), 5, 1, self.max_generations)
        max_gen.set_help(_("The number of generations " "to include in the report"))
        menu.add_option(category_name, "maxgen", max_gen)

        circle = EnumeratedListOption(_("Type of graph"), HALF_CIRCLE)
        circle.add_item(FULL_CIRCLE, _("full circle"))
        circle.add_item(HALF_CIRCLE, _("half circle"))
        circle.add_item(OVERHANG, _("overhang"))
        circle.add_item(QUAR_CIRCLE, _("quarter circle"))
        circle.set_help(
            _(
                "The form of the graph: full circle, half circle,"
                " half circle with overhang or quarter circle."
            )
        )
        menu.add_option(category_name, "circle", circle)

        background = EnumeratedListOption(_("Background color"), BACKGROUND_GEN)
        background.add_item(BACKGROUND_WHITE, _("white"))
        background.add_item(BACKGROUND_GEN, _("generation dependent"))
        background.set_help(
            _("Background color is either white or generation" " dependent")
        )
        menu.add_option(category_name, "background", background)

        radial = EnumeratedListOption(_("Orientation of radial texts"), RADIAL_UPRIGHT)
        radial.add_item(RADIAL_UPRIGHT, _("upright"))
        radial.add_item(RADIAL_ROUNDABOUT, _("roundabout"))
        radial.set_help(_("Print radial texts upright or roundabout"))
        menu.add_option(category_name, "radial", radial)
        draw_empty = BooleanOption(_("Draw empty boxes"), True)
        draw_empty.set_help(
            _("Draw the background " "although there is no information")
        )
        menu.add_option(category_name, "draw_empty", draw_empty)

        same_style = BooleanOption(_("Use one font style " "for all generations"), True)
        same_style.set_help(
            _(
                "You can customize font and color "
                "for each generation in the style editor"
            )
        )
        menu.add_option(category_name, "same_style", same_style)

        flip_text = BooleanOption(
            _("Flip names that are on the " "bottom half of the fan"), True
        )
        flip_text.set_help(
            _("Flip names for generations 2, 3 and 4 " "i.e. those found in circles")
        )
        menu.add_option(category_name, "flip_text", flip_text)

        exclude_title = BooleanOption(_("Exclude the title of the chart"), False)
        exclude_title.set_help(_("Exclude the title to make more room for the chart"))
        menu.add_option(category_name, "exclude_title", exclude_title)

        category_name = _("Report Options (2)")

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        stdoptions.add_localization_option(menu, category_name)

    def make_default_style(self, default_style):
        """Make the default output style for the Fan Chart report."""
        background_colors = [
            (255, 63, 0),
            (255, 175, 15),
            (255, 223, 87),
            (255, 255, 111),
            (159, 255, 159),
            (111, 215, 255),
            (79, 151, 255),
            (231, 23, 255),
            (231, 23, 221),
            (210, 170, 124),
            (189, 153, 112),
        ]

        # Paragraph Styles
        f_style = FontStyle()
        f_style.set_size(18)
        f_style.set_bold(1)
        f_style.set_type_face(FONT_SANS_SERIF)
        p_style = ParagraphStyle()
        p_style.set_font(f_style)
        p_style.set_alignment(PARA_ALIGN_CENTER)
        p_style.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("FC-Title", p_style)

        f_style = FontStyle()
        f_style.set_size(9)
        f_style.set_type_face(FONT_SANS_SERIF)
        p_style = ParagraphStyle()
        p_style.set_font(f_style)
        p_style.set_alignment(PARA_ALIGN_CENTER)
        p_style.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("FC-Text", p_style)

        for i in range(0, self.max_generations):
            f_style = FontStyle()
            f_style.set_size(9)
            f_style.set_type_face(FONT_SANS_SERIF)
            p_style = ParagraphStyle()
            p_style.set_font(f_style)
            p_style.set_alignment(PARA_ALIGN_CENTER)
            p_style.set_description(
                _('The style used for the text display of generation "%d"') % i
            )
            default_style.add_paragraph_style("FC-Text" + "%02d" % i, p_style)

        # GraphicsStyles
        g_style = GraphicsStyle()
        g_style.set_paragraph_style("FC-Title")
        default_style.add_draw_style("FC-Graphic-title", g_style)

        for i in range(0, self.max_generations):
            g_style = GraphicsStyle()
            g_style.set_paragraph_style("FC-Text" + "%02d" % i)
            g_style.set_fill_color(background_colors[i])
            default_style.add_draw_style("FC-Graphic" + "%02d" % i, g_style)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Italian version, 2009 (derived from the catalan version)

"""
Italian-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Italian parser
#
# -------------------------------------------------------------------------
class DateParserIT(DateParser):
    modifier_to_int = {
        "prima del": Date.MOD_BEFORE,
        "prima": Date.MOD_BEFORE,
        "dopo del": Date.MOD_AFTER,
        "dopo": Date.MOD_AFTER,
        "approssimativamente": Date.MOD_ABOUT,
        "apross.": Date.MOD_ABOUT,
        "apross": Date.MOD_ABOUT,
        "circa il": Date.MOD_ABOUT,
        "circa": Date.MOD_ABOUT,
        "ca.": Date.MOD_ABOUT,
        "ca": Date.MOD_ABOUT,
        "c.": Date.MOD_ABOUT,
        "da": Date.MOD_FROM,
        "a": Date.MOD_TO,
    }

    calendar_to_int = {
        "gregoriano": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "giuliano": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "ebraico": Date.CAL_HEBREW,
        "e": Date.CAL_HEBREW,
        "islamico": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "rivoluzionario": Date.CAL_FRENCH,
        "r": Date.CAL_FRENCH,
        "persiano": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "svedese": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "stimata": Date.QUAL_ESTIMATED,
        "st.": Date.QUAL_ESTIMATED,
        "st": Date.QUAL_ESTIMATED,
        "calcolata": Date.QUAL_CALCULATED,
        "calc.": Date.QUAL_CALCULATED,
        "calc": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        DateParser.init_strings(self)
        _span_1 = ["dal", "da"]
        _span_2 = ["al", "a"]
        _range_1 = ["tra", "fra"]
        _range_2 = ["e"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Italian display
#
# -------------------------------------------------------------------------
class DateDisplayIT(DateDisplay):
    """
    Italian language date display class.
    """

    long_months = (
        "",
        "gennaio",
        "febbraio",
        "marzo",
        "aprile",
        "maggio",
        "giugno",
        "luglio",
        "agosto",
        "settembre",
        "ottobre",
        "novembre",
        "dicembre",
    )

    short_months = (
        "",
        "gen",
        "feb",
        "mar",
        "apr",
        "mag",
        "giu",
        "lug",
        "ago",
        "set",
        "ott",
        "nov",
        "dic",
    )

    calendar = (
        "",
        "Giuliano",
        "Ebraico",
        "Rivoluzionario",
        "Persiano",
        "Islamico",
        "Svedese",
    )

    _mod_str = ("", "prima del ", "dopo del ", "circa il ", "", "", "")

    _qual_str = ("", "stimata ", "calcolata ")

    french = (
        "",
        "vendemmiaio",
        "brumaio",
        "frimaio",
        "nevoso",
        "piovoso",
        "ventoso",
        "germile",
        "fiorile",
        "pratile",
        "messidoro",
        "termidoro",
        "fruttidoro",
        "extra",
    )

    formats = (
        "AAAA-MM-DD (ISO)",
        "Numerico",
        "Mese Giorno Anno",
        "MES Giorno, Anno",
        "Giorno Mese Anno",
        "Giorno MES Anno",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "dal", d1, "al", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "tra", d1, "e", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("it_IT", "it", "italian", "Italian", "it_CH", ("%d/%m/%Y",)),
    DateParserIT,
    DateDisplayIT,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Place object for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .primaryobj import PrimaryObject
from .placeref import PlaceRef
from .placename import PlaceName
from .placetype import PlaceType
from .citationbase import CitationBase
from .notebase import NoteBase
from .mediabase import MediaBase
from .urlbase import UrlBase
from .tagbase import TagBase
from .location import Location
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Place class
#
# -------------------------------------------------------------------------
class Place(CitationBase, NoteBase, MediaBase, UrlBase, PrimaryObject):
    """
    Contains information related to a place, including multiple address
    information (since place names can change with time), longitude, latitude,
    a collection of images and URLs, a note and a source.
    """

    def __init__(self, source=None):
        """
        Create a new Place object, copying from the source if present.

        :param source: A Place object used to initialize the new Place
        :type source: Place
        """
        PrimaryObject.__init__(self, source)
        CitationBase.__init__(self, source)
        NoteBase.__init__(self, source)
        MediaBase.__init__(self, source)
        UrlBase.__init__(self, source)
        if source:
            self.long = source.long
            self.lat = source.lat
            self.title = source.title
            self.name = source.name
            self.alt_names = source.alt_names
            self.placeref_list = list(map(PlaceRef, source.placeref_list))
            self.place_type = source.place_type
            self.code = source.code
            self.alt_loc = list(map(Location, source.alt_loc))
        else:
            self.long = ""
            self.lat = ""
            self.title = ""
            self.name = PlaceName()
            self.alt_names = []
            self.placeref_list = []
            self.place_type = PlaceType()
            self.code = ""
            self.alt_loc = []

    def serialize(self):
        """
        Convert the data held in the Place to a Python tuple that
        represents all the data elements.

        This method is used to convert the object into a form that can easily
        be saved to a database.

        These elements may be primitive Python types (string, integers),
        complex Python types (lists or tuples, or Python objects. If the
        target database cannot handle complex types (such as objects or
        lists), the database is responsible for converting the data into
        a form that it can use.

        :returns: Returns a python tuple containing the data that should
                  be considered persistent.
        :rtype: tuple
        """
        return (
            self.handle,
            self.gramps_id,
            self.title,
            self.long,
            self.lat,
            [pr.serialize() for pr in self.placeref_list],
            self.name.serialize(),
            [an.serialize() for an in self.alt_names],
            self.place_type.serialize(),
            self.code,
            [al.serialize() for al in self.alt_loc],
            UrlBase.serialize(self),
            MediaBase.serialize(self),
            CitationBase.serialize(self),
            NoteBase.serialize(self),
            self.change,
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .url import Url
        from .mediaref import MediaRef

        return {
            "type": "object",
            "title": _("Place"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {"type": "string", "maxLength": 50, "title": _("Handle")},
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "title": {"type": "string", "title": _("Title")},
                "long": {"type": "string", "title": _("Longitude")},
                "lat": {"type": "string", "title": _("Latitude")},
                "placeref_list": {
                    "type": "array",
                    "items": PlaceRef.get_schema(),
                    "title": _("Places"),
                },
                "name": PlaceName.get_schema(),
                "alt_names": {
                    "type": "array",
                    "items": PlaceName.get_schema(),
                    "title": _("Alternate Names"),
                },
                "place_type": PlaceType.get_schema(),
                "code": {"type": "string", "title": _("Code")},
                "alt_loc": {
                    "type": "array",
                    "items": Location.get_schema(),
                    "title": _("Alternate Locations"),
                },
                "urls": {
                    "type": "array",
                    "items": Url.get_schema(),
                    "title": _("URLs"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Place object.

        :param data: tuple containing the persistent data associated with the
                     Place object
        :type data: tuple
        """
        (
            self.handle,
            self.gramps_id,
            self.title,
            self.long,
            self.lat,
            placeref_list,
            name,
            alt_names,
            the_type,
            self.code,
            alt_loc,
            urls,
            media_list,
            citation_list,
            note_list,
            self.change,
            tag_list,
            self.private,
        ) = data

        self.place_type = PlaceType()
        self.place_type.unserialize(the_type)
        self.alt_loc = [Location().unserialize(al) for al in alt_loc]
        self.placeref_list = [PlaceRef().unserialize(pr) for pr in placeref_list]
        self.name = PlaceName().unserialize(name)
        self.alt_names = [PlaceName().unserialize(an) for an in alt_names]
        UrlBase.unserialize(self, urls)
        MediaBase.unserialize(self, media_list)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        TagBase.unserialize(self, tag_list)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.long, self.lat, self.title, self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """

        ret = self.media_list + self.alt_loc + self.urls + [self.name] + self.alt_names
        return ret

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: List of child secondary child objects that may refer citations.
        :rtype: list
        """
        return self.media_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                refer notes.
        :rtype: list
        """
        return self.media_list

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list() + self.placeref_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return (
            self.get_referenced_note_handles()
            + self.get_referenced_citation_handles()
            + self.get_referenced_tag_handles()
        )

    def merge(self, acquisition):
        """Merge the content of acquisition into this place.

        :param acquisition: The place to merge with the present place.
        :type acquisition: Place
        """
        self._merge_privacy(acquisition)
        self._merge_locations(acquisition)
        self._merge_alt_names(acquisition)
        self._merge_media_list(acquisition)
        self._merge_url_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_tag_list(acquisition)
        self._merge_placeref_list(acquisition)

    def set_title(self, title):
        """
        Set the descriptive title of the Place object.

        :param title: descriptive title to assign to the Place
        :type title: str
        """
        self.title = title

    def get_title(self):
        """
        Return the descriptive title of the Place object.

        :returns: Returns the descriptive title of the Place
        :rtype: str
        """
        return self.title

    def set_name(self, name):
        """
        Set the name of the Place object.

        :param name: name to assign to the Place
        :type name: PlaceName
        """
        if not isinstance(name, PlaceName):
            raise ValueError("Place.set_name(name) requires a PlaceName()")
        self.name = name

    def get_name(self):
        """
        Return the name of the Place object.

        :returns: Returns the name of the Place
        :rtype: PlaceName
        """
        return self.name

    def get_all_names(self):
        """
        Return a list of all names of the Place object.

        :returns: Returns a list of all names of the Place
        :rtype: list of PlaceName
        """
        return [self.name] + self.alt_names

    def set_longitude(self, longitude):
        """
        Set the longitude of the Place object.

        :param longitude: longitude to assign to the Place
        :type longitude: str
        """
        self.long = longitude

    def get_longitude(self):
        """
        Return the longitude of the Place object.

        :returns: Returns the longitude of the Place
        :rtype: str
        """
        return self.long

    def set_latitude(self, latitude):
        """
        Set the latitude of the Place object.

        :param latitude: latitude to assign to the Place
        :type latitude: str
        """
        self.lat = latitude

    def get_latitude(self):
        """
        Return the latitude of the Place object.

        :returns: Returns the latitude of the Place
        :rtype: str
        """
        return self.lat

    def set_type(self, place_type):
        """
        Set the type of the Place object.

        :param type: type to assign to the Place
        :type type: PlaceType
        """
        self.place_type.set(place_type)

    def get_type(self):
        """
        Return the type of the Place object.

        :returns: Returns the type of the Place
        :rtype: PlaceType
        """
        return self.place_type

    def set_code(self, code):
        """
        Set the code of the Place object.

        :param code: code to assign to the Place
        :type code: str
        """
        self.code = code

    def get_code(self):
        """
        Return the code of the Place object.

        :returns: Returns the code of the Place
        :rtype: str
        """
        return self.code

    def add_placeref(self, placeref):
        """
        Add a place reference to the list of place references.

        :param code: place reference to append to the list
        :type code: PlaceRef
        """
        self.placeref_list.append(placeref)

    def get_placeref_list(self):
        """
        Return the place reference list for the Place object.

        :returns: Returns the place reference list for the Place
        :rtype: list
        """
        return self.placeref_list

    def set_placeref_list(self, placeref_list):
        """
        Set the place reference list for the Place object.

        :param code: place reference list to assign to the Place
        :type code: list
        """
        self.placeref_list = placeref_list

    def _merge_placeref_list(self, acquisition):
        """
        Add the main and alternate locations of acquisition to the alternate
        location list.

        :param acquisition: instance to merge
        :type acquisition: :class:'~.place.Place
        """
        placeref_list = self.placeref_list[:]
        add_list = acquisition.placeref_list
        for addendum in add_list:
            for placeref in placeref_list:
                if placeref.is_equal(addendum):
                    break
            else:
                self.placeref_list.append(addendum)

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                this object type.
        :rtype: bool
        """
        if classname == "Place":
            for placeref in self.placeref_list:
                if placeref.ref == handle:
                    return True
        return False

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Place":
            for placeref in self.placeref_list:
                if placeref.ref == old_handle:
                    placeref.ref = new_handle

    def get_alternative_names(self):
        """
        Return a list of alternative names for the current Place.

        :returns: Returns the alternative names for the Place
        :rtype: list of PlaceName
        """
        return self.alt_names

    def set_alternative_names(self, name_list):
        """
        Replace the current alternative names list with the new one.

        :param name_list: The list of names to assign to the Place's internal
                          list.
        :type name_list: list of PlaceName
        """
        self.alt_names = name_list

    def add_alternative_name(self, name):
        """
        Add a name to the alternative names list.

        :param name: name to add
        :type name: string
        """
        if name not in self.alt_names:
            self.alt_names.append(name)

    def get_alternate_locations(self):
        """
        Return a list of alternate :class:`~.location.Location` objects the
        present alternate information about the current Place.

        A Place can have more than one :class:`~.location.Location`, since
        names and jurisdictions can change over time for the same place.

        :returns: Returns the alternate :class:`~.location.Location` objects
                  for the Place
        :rtype: list of :class:`~.location.Location` objects
        """
        return self.alt_loc

    def set_alternate_locations(self, location_list):
        """
        Replace the current alternate :class:`~.location.Location` object list
        with the new one.

        :param location_list: The list of :class:`~.location.Location` objects
                              to assign to the Place's internal list.
        :type location_list: list of :class:`~.location.Location` objects
        """
        self.alt_loc = location_list

    def add_alternate_locations(self, location):
        """
        Add a :class:`~.location.Location` object to the alternate location
        list.

        :param location: :class:`~.location.Location` instance to add
        :type location: :class:`~.location.Location`
        """
        if location not in self.alt_loc:
            self.alt_loc.append(location)

    def _merge_locations(self, acquisition):
        """
        Add the main and alternate locations of acquisition to the alternate
        location list.

        :param acquisition: instance to merge
        :type acquisition: :class:'~.place.Place
        """
        altloc_list = self.alt_loc[:]
        add_list = acquisition.get_alternate_locations()
        for addendum in add_list:
            for altloc in altloc_list:
                if altloc.is_equal(addendum):
                    break
            else:
                self.alt_loc.append(addendum)

    def _merge_alt_names(self, acquisition):
        """
        Add the main and alternative names of acquisition to the alternative
        names list.

        :param acquisition: instance to merge
        :type acquisition: :class:'~.place.Place
        """
        if acquisition.name.value:
            if acquisition.name != self.name:
                if acquisition.name not in self.alt_names:
                    self.alt_names.append(acquisition.name)

        for addendum in acquisition.alt_names:
            if addendum.value:
                if addendum != self.name:
                    if addendum not in self.alt_names:
                        self.alt_names.append(addendum)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2012       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the Plugin class for document generator plugins.
"""

from . import Plugin
from .docgen import TextDoc, DrawDoc


class DocGenPlugin(Plugin):
    """
    This class represents a plugin for generating documents from Gramps
    """

    def __init__(
        self,
        name,
        description,
        basedoc,
        paper,
        style,
        extension,
        docoptclass,
        basedocname,
    ):
        """
        :param name: A friendly name to call this plugin.
            Example: "Plain Text"
        :type name: string
        :param description: A short description of the plugin.
            Example: "This plugin will generate text documents in plain text."
        :type description: string
        :param basedoc: A class that implements the BaseDoc
            interface.
        :type basedoc: BaseDoc
        :param paper: Indicates whether the plugin uses paper or not.
            True = use paper; False = do not use paper
        :type paper: bool
        :param style: Indicates whether the plugin uses styles or not.
            True = use styles; False = do not use styles
        :type style: bool
        :param extension: The extension for the output file.
            Example: "txt"
        :type extension: str
        :param docoptclass: either None or a subclass of DocOptions
        :type docoptclass: either None or a DocOptions subclass
        :param basedocname: The BaseDoc name of this plugin.
            Example: "AsciiDoc"
        :type basedocname: string
        :return: nothing
        """
        Plugin.__init__(self, name, description, basedoc.__module__)
        self.__basedoc = basedoc
        self.__paper = paper
        self.__style = style
        self.__extension = extension
        self.__docoptclass = docoptclass
        self.__basedocname = basedocname

    def get_basedoc(self):
        """
        Get the :class:`.BaseDoc` class for this plugin.

        :return: the :class:`.BaseDoc` class passed into :meth:`__init__`
        """
        return self.__basedoc

    def get_paper_used(self):
        """
        Get the paper flag for this plugin.

        :return: bool - True = use paper; False = do not use paper
        """
        return self.__paper

    def get_style_support(self):
        """
        Get the style flag for this plugin.

        :return: bool - True = use styles; False = do not use styles
        """
        return self.__style

    def get_extension(self):
        """
        Get the file extension for the output file.

        :return: str
        """
        return self.__extension

    def get_doc_option_class(self):
        """
        Get the :class:`.DocOptions` subclass for this plugin, if any

        :return: the :class:`.DocOptions` subclass passed into :meth:`__init__`
        """
        return self.__docoptclass

    def get_basedocname(self):
        """
        Get the :class:`.BaseDoc` name for this plugin.

        :return: the :class:`.BaseDoc` name passed into :meth:`__init__`
        """
        return self.__basedocname

    def get_text_support(self):
        """
        Check if the plugin supports the :class:`.TextDoc` interface.

        :return: bool: True if :class:`.TextDoc` is supported; False if
                       :class:`.TextDoc` is not supported.
        """
        return bool(issubclass(self.__basedoc, TextDoc))

    def get_draw_support(self):
        """
        Check if the plugin supports the :class:`.DrawDoc` interface.

        :return: bool: True if :class:`.DrawDoc` is supported; False if
                       :class:`.DrawDoc` is not supported.
        """
        return bool(issubclass(self.__basedoc, DrawDoc))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base type for all gramps types.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

_UNKNOWN = _("Unknown")


# -------------------------------------------------------------------------
#
# GrampsTypeMeta class
#
# -------------------------------------------------------------------------
class GrampsTypeMeta(type):
    """
    Metaclass for :class:`~.grampstype.GrampsType`.

    Create the class-specific integer/string maps.
    """

    def __init__(cls, name, bases, namespace):
        # Helper function to create the maps
        def init_map(data, key_col, data_col, blacklist=None):
            """
            Initialize the map, building a new map from the specified
            columns.
            """
            if blacklist:
                return dict(
                    [
                        (item[key_col], item[data_col])
                        for item in data
                        if item[0] not in blacklist
                    ]
                )
            else:
                return dict([(item[key_col], item[data_col]) for item in data])

        # Call superclass initialization
        type.__init__(cls, name, bases, namespace)

        # Build the integer/string maps
        if hasattr(cls, "_DATAMAP"):
            cls._I2SMAP = init_map(cls._DATAMAP, 0, 1, cls._BLACKLIST)
            cls._S2IMAP = init_map(cls._DATAMAP, 1, 0, cls._BLACKLIST)
            cls._I2EMAP = init_map(cls._DATAMAP, 0, 2, cls._BLACKLIST)
            cls._E2IMAP = init_map(cls._DATAMAP, 2, 0, cls._BLACKLIST)


# -------------------------------------------------------------------------
#
# GrampsType class
#
# -------------------------------------------------------------------------
class GrampsType(object, metaclass=GrampsTypeMeta):
    """Base class for all Gramps object types.

    :cvar _DATAMAP:
      (list) 3-tuple like (index, localized_string, english_string).
    :cvar _BLACKLIST:
      List of indices to ignore (obsolete/retired entries).
      (gramps policy is never to delete type values, or reuse the name (TOKEN)
      of any specific type value)
    :cvar POS_<x>: (int)
      Position of <x> attribute in the serialized format of
      an instance.

    .. warning:: The POS_<x> class variables reflect the serialized object,
                 they have to be updated in case the data structure or the
                 :meth:`serialize` method changes!

    :cvar _CUSTOM:  (int) a custom type object
    :cvar _DEFAULT: (int) the default type, used on creation

    :attribute value: (int) Returns or sets integer value
    :attribute string: (str) Returns or sets string value
    """

    (POS_VALUE, POS_STRING) = list(range(2))

    _CUSTOM = 0
    _DEFAULT = 0

    _DATAMAP = []
    _BLACKLIST = None
    _I2SMAP = {}
    _S2IMAP = {}
    _I2EMAP = {}
    _E2IMAP = {}
    _MENU = []
    __slots__ = ("__value", "__string")

    def __getstate__(self):
        return {"__value": self.__value, "__string": self.__string}

    def __setstate__(self, dict_):
        self.__value = dict_["__value"]
        if self.__value == self._CUSTOM:
            self.__string = dict_["__string"]
        else:
            self.__string = ""

    def __init__(self, value=None):
        """
        Create a new type, initialize the value from one of several possible
        states.
        """
        self.__value = self._DEFAULT
        self.__string = ""
        if value is not None:
            self.set(value)

    def __set_tuple(self, value):
        "Set the value/string properties from a tuple."
        val, strg = self._DEFAULT, ""
        if value:
            val = value[0]
            if len(value) > 1 and val == self._CUSTOM:
                strg = value[1]
        self.__value = val
        self.__string = strg

    def __set_int(self, value):
        "Set the value/string properties from an integer."
        self.__value = value
        self.__string = ""

    def __set_instance(self, value):
        "Set the value/string properties from another grampstype."
        self.__value = value.value
        if self.__value == self._CUSTOM:
            self.__string = value.string
        else:
            self.__string = ""

    def __set_str(self, value):
        "Set the value/string properties from a string."
        self.__value = self._S2IMAP.get(value, self._CUSTOM)
        if self.__value == self._CUSTOM:
            self.__string = value
        else:
            self.__string = ""

    def set(self, value):
        "Set the value/string properties from the passed in value."
        if isinstance(value, tuple):
            self.__set_tuple(value)
        elif isinstance(value, int):
            self.__set_int(value)
        elif isinstance(value, self.__class__):
            self.__set_instance(value)
        elif isinstance(value, str):
            self.__set_str(value)
        else:
            self.__value = self._DEFAULT
            self.__string = ""

    def set_from_xml_str(self, value):
        """
        This method sets the type instance based on the untranslated string
        (obtained e.g. from XML).
        """
        if value in self._E2IMAP:
            self.__value = self._E2IMAP[value]
            self.__string = ""
            if self.__value == self._CUSTOM:
                # if the custom event is actually 'Custom' then we should save it
                # with that string value. That is, 'Custom' is in _E2IMAP
                self.__string = value
        else:
            self.__value = self._CUSTOM
            self.__string = value

    def xml_str(self):
        """
        Return the untranslated string (e.g. suitable for XML) corresponding
        to the type.
        """
        if self.__value == self._CUSTOM:
            return self.__string
        elif self.__value in self._I2EMAP:
            return self._I2EMAP[self.__value]
        else:
            return _UNKNOWN

    def serialize(self):
        """Convert the object to a serialized tuple of data."""
        return (self.__value, self.__string)

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Type"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "string": {"type": "string", "title": _("Type")},
            },
        }

    def unserialize(self, data):
        """Convert a serialized tuple of data to an object."""
        self.__value, self.__string = data
        if self.__value != self._CUSTOM:
            self.__string = ""
        return self

    def __str__(self):
        if self.__value == self._CUSTOM:
            return self.__string
        else:
            return self._I2SMAP.get(self.__value, _UNKNOWN)

    def __int__(self):
        return self.__value

    def get_map(self):
        return self._I2SMAP

    def get_standard_names(self):
        """Return the list of localized names for all standard types."""
        return [
            s
            for (i, s) in list(self._I2SMAP.items())
            if (i != self._CUSTOM) and s.strip()
        ]

    def get_standard_xml(self):
        """Return the list of XML (english) names for all standard types."""
        return [
            s
            for (i, s) in list(self._I2EMAP.items())
            if (i != self._CUSTOM) and s.strip()
        ]

    def is_custom(self):
        return self.__value == self._CUSTOM

    def is_default(self):
        return self.__value == self._DEFAULT

    def get_custom(self):
        return self._CUSTOM

    def get_menu(self):
        """Return the list of localized names for the menu."""
        if self._MENU:
            return [[_(i), s] for (i, s) in self._MENU]
        return self._MENU

    def get_menu_standard_xml(self):
        """Return the list of XML (english) names for the menu."""
        return self._MENU

    def __eq__(self, value):
        if isinstance(value, int):
            return self.__value == value
        elif isinstance(value, str):
            if self.__value == self._CUSTOM:
                return self.__string == value
            else:
                return self._I2SMAP.get(self.__value) == value
        elif isinstance(value, tuple):
            if self.__value == self._CUSTOM:
                return (self.__value, self.__string) == value
            else:
                return self.__value == value[0]
        else:
            if value.value == self._CUSTOM and self.__value == self._CUSTOM:
                return self.__string == value.string
            elif value.value != self._CUSTOM and self.__value != self._CUSTOM:
                return self.__value == value.value
            else:
                return False

    def __ne__(self, value):
        return not self.__eq__(value)

    value = property(__int__, __set_int, None, "Returns or sets integer value")
    string = property(__str__, __set_str, None, "Returns or sets string value")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject


# -------------------------------------------------------------------------
#
# This is used by plugins to create a menu of available filters
#
# -------------------------------------------------------------------------
def build_filter_model(space, local=[]):
    from gramps.gen.filters import CustomFilters

    model = Gtk.ListStore(GObject.TYPE_STRING, object)

    if isinstance(space, str):
        flist = local + CustomFilters.get_filters(space)
    elif isinstance(space, (list, tuple)):
        flist = space
    else:
        return None

    for filt in flist:
        model.append(row=[filt.get_name(), filt])
    return model



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# ------------------------------------------------------------------------
#
# Register Gramplet
#
# ------------------------------------------------------------------------
class FAQGramplet(Gramplet):
    def init(self):
        self.set_use_markup(True)
        self.clear_text()
        """
        The "wiki" HTTP "anchor type" is gramps-specific, and causes
        GuiGramplet's on_button_press method to invoke gui/display.py's
        display_help method, which prepends URL_WIKISTRING before it then
        calls gui/display.py's display_url method.  So no URL_WIKISTRING
        is needed in this code.
        """
        faq_dict = {
            "bold_start": "<b>",
            "bold_end": "</b>",
            "gramps_wiki_html_start": "<a wiki='",
            "gramps_manual_html_start": "<a wiki='" + URL_MANUAL_PAGE + "_-_",
            "gramps_FAQ_html_start": "<a wiki='" + URL_MANUAL_PAGE + "_-_FAQ",
            "html_middle": "'>",
            "html_end": "</a>",
        }
        self.render_text(
            _(
                "%(bold_start)s%(gramps_FAQ_html_start)s%(html_middle)s"
                "Frequently Asked Questions"
                "%(html_end)s%(bold_end)s"
                "\n(needs a connection to the internet)\n"
            )
            % faq_dict
        )

        self.render_text("\n<b>%s</b>\n\n" % _("Editing Spouses"))

        faq_dict.update({"faq_section": "#How_do_I_change_the_order_of_spouses.3F'>"})
        self.render_text(
            _(
                "  1. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do I change the order of spouses?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"faq_section": "#How_do_I_add_an_additional_spouse.3F'>"})
        self.render_text(
            _(
                "  2. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do I add an additional spouse?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"faq_section": "#How_do_I_remove_a_spouse.3F'>"})
        self.render_text(
            _(
                "  3. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do I remove a spouse?"
                "%(html_end)s\n"
            )
            % faq_dict
        )

        self.render_text("\n<b>%s</b>\n\n" % _("Backups and Updates"))

        faq_dict.update({"faq_section": "#How_do_I_keep_backups.3F'>"})
        self.render_text(
            _(
                "  4. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do I make backups safely?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"faq_section": "#How_do_I_upgrade_GRAMPS.3F'>"})
        self.render_text(
            _(
                "  5. %(gramps_FAQ_html_start)s%(faq_section)s"
                "Is it necessary to update Gramps "
                "every time an update is released?"
                "%(html_end)s\n"
            )
            % faq_dict
        )

        self.render_text("\n<b>%s</b>\n\n" % _("Data Entry"))

        faq_dict.update(
            {
                "section": "Entering_and_Editing_Data:_Detailed_-_part_1"
                "#Editing_Information_About_Relationships'>"
            }
        )
        self.render_text(
            _(
                "  6. %(gramps_manual_html_start)s%(section)s"
                "How should information about marriages be entered?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update(
            {
                "faq_section": "#What_is_the_difference_"
                "between_a_residence_and_an_address.3F'>"
            }
        )
        self.render_text(
            _(
                "  7. %(gramps_FAQ_html_start)s%(faq_section)s"
                "What's the difference between a residence and an address?"
                "%(html_end)s\n"
            )
            % faq_dict
        )

        self.render_text("\n<b>%s</b>\n\n" % _("Media Files"))

        faq_dict.update({"faq_section": "#How_do_you_add_photos_to_an_item.3F'>"})
        self.render_text(
            _(
                "  8. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do you add a photo of a person/source/event?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"faq_section": "#How_do_you_find_unused_media.3F'>"})
        self.render_text(
            _(
                "  9. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How do you find unused media objects?"
                "%(html_end)s\n"
            )
            % faq_dict
        )

        self.render_text("\n<b>%s</b>\n\n" % _("Miscellaneous"))

        faq_dict.update(
            {"faq_section": "#How_can_I_publish_web_sites_" "generated_by_GRAMPS.3F'>"}
        )
        self.render_text(
            _(
                " 10. %(gramps_FAQ_html_start)s%(faq_section)s"
                "How can I make a website with Gramps and my tree?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update(
            {
                "web_html_start": "<a href='http://sourceforge.net/mailarchive"
                "/message.php?msg_id=21487967'>"
            }
        )
        self.render_text(
            _(
                " 11. %(web_html_start)s"
                "How do I record one's occupation?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"faq_section": "#What_do_I_do_if_I_have_found_a_bug.3F'>"})
        self.render_text(
            _(
                " 12. %(gramps_FAQ_html_start)s%(faq_section)s"
                "What do I do if I have found a bug?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"section": "Portal:Using_GRAMPS'>"})
        self.render_text(
            _(
                " 13. %(gramps_wiki_html_start)s%(section)s"
                "Is there a manual for Gramps?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"section": "Category:Tutorials'>"})
        self.render_text(
            _(
                " 14. %(gramps_wiki_html_start)s%(section)s"
                "Are there tutorials available?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"section": "Category:How_do_I...'>"})
        self.render_text(
            _(
                " 15. %(gramps_wiki_html_start)s%(section)s"
                "How do I ...?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        faq_dict.update({"section": "How_you_can_help'>"})
        self.render_text(
            _(
                " 16. %(gramps_wiki_html_start)s%(section)s"
                "How can I help with Gramps?"
                "%(html_end)s\n"
            )
            % faq_dict
        )
        self.append_text("", scroll_to="begin")
